"""Regression tests for the embedded Juvenal API."""

from __future__ import annotations

import json
import subprocess
from pathlib import Path
from unittest.mock import patch

import pytest
import yaml

import juvenal.api as api_module
from juvenal.api import (
    JuvenalExecutionError,
    JuvenalUsageError,
    do,
    goal,
    plan_and_do,
)
from juvenal.engine import PlanResult
from juvenal.state import PipelineState
from juvenal.workflow import load_workflow as load_workflow_impl
from tests.conftest import MockBackend


def _init_git_repo(path: Path) -> None:
    path.mkdir()
    subprocess.run(["git", "init"], cwd=path, check=True, capture_output=True, text=True)


def _write_planned_workflow_structure(base_dir: Path, phases: list[dict[str, object]] | None = None) -> None:
    plan_dir = base_dir / ".plan"
    plan_dir.mkdir(exist_ok=True)
    payload = {
        "linear": True,
        "yaml_source_mode": "inline-only",
        "verifier_encoding": "explicit-phases",
        "phases": phases
        if phases is not None
        else [
            {
                "order": 1,
                "id": "execute",
                "type": "implement",
                "bounce_target": None,
                "required_preexisting_inputs": [],
            }
        ],
    }
    (plan_dir / "workflow-structure.yaml").write_text(yaml.safe_dump(payload, sort_keys=False), encoding="utf-8")


def _write_planned_workflow(
    base_dir: Path,
    workflow_text: str,
    *,
    phases: list[dict[str, object]] | None = None,
) -> Path:
    workflow_path = (base_dir / "workflow.yaml").resolve()
    workflow_path.write_text(workflow_text, encoding="utf-8")
    _write_planned_workflow_structure(base_dir, phases=phases)
    return workflow_path


def test_goal_rejects_missing_working_dir(tmp_path):
    with pytest.raises(JuvenalUsageError, match="Working directory does not exist"):
        with goal("Goal", working_dir=tmp_path / "missing", backend=MockBackend()):
            pass


def test_goal_rejects_non_directory_working_dir(tmp_path):
    file_path = tmp_path / "goal.txt"
    file_path.write_text("not a directory")

    with pytest.raises(JuvenalUsageError, match="Working directory is not a directory"):
        with goal("Goal", working_dir=file_path, backend=MockBackend()):
            pass


def test_goal_rejects_unknown_backend(tmp_path):
    with pytest.raises(JuvenalUsageError, match="Unknown backend"):
        with goal("Goal", working_dir=tmp_path, backend="not-a-backend"):
            pass


def test_goal_restores_nested_session_and_allocates_distinct_default_session_dirs(tmp_path):
    outer_backend = MockBackend()
    inner_backend = MockBackend()
    outer_backend.add_response(exit_code=0, output="outer done")
    inner_backend.add_response(exit_code=0, output="inner done")

    with goal("Outer goal", working_dir=tmp_path, backend=outer_backend) as outer:
        assert outer.artifact_root == (tmp_path / ".juvenal-api").resolve()
        assert outer.session_artifact_dir == (tmp_path / ".juvenal-api" / "session-001").resolve()
        assert outer.session_artifact_dir.is_dir()

        with goal("Inner goal", working_dir=tmp_path, backend=inner_backend) as inner:
            assert inner.artifact_root == outer.artifact_root
            assert inner.session_artifact_dir == (tmp_path / ".juvenal-api" / "session-002").resolve()
            assert inner.session_artifact_dir.is_dir()
            assert inner.session_artifact_dir != outer.session_artifact_dir
            do("Inner task")
            assert [entry["instruction"] for entry in inner.history] == ["Inner task"]

        do("Outer task")
        assert [entry["instruction"] for entry in outer.history] == ["Outer task"]


@pytest.mark.parametrize(
    "session_name",
    ["", "Example", "has_underscore", "two--parts", "-leading", "trailing-", "with space", " spaced", "a/b", ".", ".."],
)
def test_goal_rejects_invalid_session_names(tmp_path, session_name):
    with pytest.raises(JuvenalUsageError, match="session_name must match"):
        with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name=session_name):
            pass


def test_goal_rejects_reserved_named_session_namespace(tmp_path):
    with pytest.raises(JuvenalUsageError, match="reserved for anonymous sessions"):
        with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name="session-001"):
            pass


@pytest.mark.parametrize("manifest_text", [None, "{not json"])
def test_goal_named_session_requires_valid_manifest_when_directory_exists(tmp_path, manifest_text):
    session_dir = tmp_path / ".juvenal-api" / "example"
    session_dir.mkdir(parents=True)
    if manifest_text is not None:
        (session_dir / "session.json").write_text(manifest_text)

    with pytest.raises(JuvenalUsageError, match="Invalid session manifest"):
        with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
            pass


def test_goal_named_session_rejects_manifest_missing_session_name(tmp_path):
    with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        manifest = json.loads(session.manifest_path.read_text())

    del manifest["session_name"]
    session.manifest_path.write_text(json.dumps(manifest))

    with pytest.raises(JuvenalUsageError, match="session_name is required for named sessions"):
        with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
            pass


def test_goal_named_session_rejects_identity_mismatches(tmp_path):
    working_dir = tmp_path / "work"
    working_dir.mkdir()
    other_working_dir = tmp_path / "other"
    other_working_dir.mkdir()
    artifact_dir = tmp_path / "artifacts"

    class AlternateBackend(MockBackend):
        def name(self) -> str:
            return "alternate-mock"

    with goal(
        "Goal",
        working_dir=working_dir,
        artifact_dir=artifact_dir,
        backend=MockBackend(),
        session_name="example",
    ):
        pass

    mismatch_cases = [
        ({"description": "Different goal"}, "goal_text"),
        ({"working_dir": other_working_dir}, "working_dir"),
        ({"backend": AlternateBackend()}, "backend_name"),
        ({"max_bounces": 1}, "max_bounces"),
        ({"serialize": True}, "serialize"),
        ({"clear_context_on_bounce": True}, "clear_context_on_bounce"),
    ]

    for overrides, field_name in mismatch_cases:
        kwargs = {
            "description": "Goal",
            "working_dir": working_dir,
            "artifact_dir": artifact_dir,
            "backend": MockBackend(),
            "session_name": "example",
        }
        kwargs.update(overrides)
        with pytest.raises(JuvenalUsageError, match=field_name):
            with goal(**kwargs):
                pass


def test_goal_named_session_reuses_manifest_and_restores_state(tmp_path):
    first_backend = MockBackend()
    first_backend.add_response(exit_code=0, output="repo prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=first_backend, session_name="example") as session:
        assert session.session_id == "example"
        assert session.session_key == "example"
        assert session.session_name == "example"
        assert session.session_artifact_dir == (tmp_path / ".juvenal-api" / "example").resolve()
        assert session.manifest_path == (tmp_path / ".juvenal-api" / "example" / "session.json").resolve()
        assert session.manifest_path.exists()

        do("Prepare the repository")

        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["run_counter"] == 1
        assert [entry["instruction"] for entry in manifest["history"]] == ["Prepare the repository"]
        assert manifest["stages"]["do-001"]["status"] == "completed"

    second_backend = MockBackend()
    second_backend.add_response(exit_code=0, output="api implemented")

    with goal(
        "Ship the API",
        working_dir=tmp_path,
        backend=second_backend,
        plain=True,
        session_name="example",
    ) as session:
        assert session.plain is True
        assert session.run_counter == 1
        assert [entry["instruction"] for entry in session.history] == ["Prepare the repository"]
        assert session.stages["do-001"]["status"] == "completed"

        do("Implement the API")

        assert [entry["instruction"] for entry in session.history] == ["Prepare the repository", "Implement the API"]
        assert (session.session_artifact_dir / "run-002-do.json").exists()

        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["run_counter"] == 2
        assert [entry["instruction"] for entry in manifest["history"]] == [
            "Prepare the repository",
            "Implement the API",
        ]
        assert manifest["stages"]["do-002"]["status"] == "completed"


def test_do_requires_active_goal_session():
    with pytest.raises(JuvenalUsageError, match="requires an active juvenal.goal"):
        do("Build the feature")


@pytest.mark.parametrize(
    ("operation", "match"),
    [
        (lambda: do(""), "empty after stripping whitespace"),
        (lambda: do([]), "requires at least one task"),
        (lambda: do(["ok", 7]), "task 2 must be a string"),
        (lambda: do("Build", checker="tester", checkers=["pm"]), "either checker= or checkers="),
        (lambda: do("Build", checkers="tester"), "must be a sequence of checker specs"),
        (lambda: do("Build", checker="not-a-valid-role"), "Invalid --checker spec"),
    ],
)
def test_do_rejects_misuse_cases(tmp_path, operation, match):
    with goal("Goal", working_dir=tmp_path, backend=MockBackend()):
        with pytest.raises(JuvenalUsageError, match=match):
            operation()


def test_do_single_step_run_records_history_and_state(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="implemented")
    backend.add_response(exit_code=0, output="VERDICT: PASS")

    with goal("Ship the API", working_dir=tmp_path, backend=backend) as session:
        do("Implement the API", checker="tester")

        assert len(backend.calls) == 2
        implement_prompt = backend.calls[0]
        assert "Ship the API" in implement_prompt
        assert str(tmp_path.resolve()) in implement_prompt
        assert "Implement the API" in implement_prompt

        assert [entry["instruction"] for entry in session.history] == ["Implement the API"]
        assert session.history[0]["success"] is True
        assert "Implement the API" in session.history[0]["summary"]
        do_state_files = list(session.session_artifact_dir.glob("*-do.json"))
        assert len(do_state_files) == 1
        assert do_state_files[0].exists()


def test_do_multi_step_run_uses_completed_steps_context_and_records_each_step(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="first done")
    backend.add_response(exit_code=0, output="VERDICT: PASS")
    backend.add_response(exit_code=0, output="second done")
    backend.add_response(exit_code=0, output="VERDICT: PASS")

    with goal("Ship the API", working_dir=tmp_path, backend=backend) as session:
        do(["Scaffold handlers", "Wire auth"], checker="tester")

        assert len(backend.calls) == 4
        first_prompt = backend.calls[0]
        second_prompt = backend.calls[2]
        assert "Scaffold handlers" in first_prompt
        assert "Wire auth" not in first_prompt
        assert "Wire auth" in second_prompt
        assert "Scaffold handlers" in second_prompt
        assert second_prompt.find("Scaffold handlers") < second_prompt.find("Wire auth")

        assert [entry["instruction"] for entry in session.history] == ["Scaffold handlers", "Wire auth"]
        assert len({entry["phase_id"] for entry in session.history}) == 2


def test_do_reuses_successful_history_in_later_prompts(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="repo prepared")
    backend.add_response(exit_code=0, output="api implemented")

    with goal("Ship the API", working_dir=tmp_path, backend=backend) as session:
        do("Prepare the repository")
        first_summary = session.history[0]["summary"]

        do("Implement the API")

        assert len(backend.calls) == 2
        assert first_summary in backend.calls[1]
        assert [entry["instruction"] for entry in session.history] == ["Prepare the repository", "Implement the API"]


def test_do_persists_run_id_checkpoint_when_workflow_validation_fails(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with pytest.raises(JuvenalUsageError, match="template variable"):
            do("Implement {{VAR}} support")

        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["run_counter"] == 1
        assert manifest["history"] == []
        assert manifest["stages"]["do-001"]["status"] == "failed"

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as resumed:
        assert resumed.run_counter == 1
        assert resumed.history == []
        assert resumed.stages["do-001"]["status"] == "failed"


def test_do_preserves_partial_success_history_when_later_step_fails(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="step 1 done")
    backend.add_response(exit_code=0, output="VERDICT: PASS")
    backend.add_response(exit_code=0, output="step 2 done")
    backend.add_response(exit_code=0, output="VERDICT: FAIL: tests are still broken")

    with goal("Ship the API", working_dir=tmp_path, backend=backend, max_bounces=1) as session:
        with pytest.raises(JuvenalExecutionError) as exc_info:
            do(["Finish setup", "Finish integration"], checker="tester")

        error = exc_info.value
        assert error.inspection_path.parent == session.session_artifact_dir
        assert error.inspection_path.name.endswith("-do.json")
        assert error.inspection_path.exists()
        assert [entry["instruction"] for entry in session.history] == ["Finish setup"]
        assert session.history[0]["phase_id"]


def test_staged_do_resumes_with_stored_prompt_snapshots_and_same_run_id(tmp_path):
    first_backend = MockBackend()
    first_backend.add_response(exit_code=0, output="step 1 done")
    first_backend.add_response(exit_code=0, output="VERDICT: PASS")
    first_backend.add_response(exit_code=1, output="step 2 crashed")

    with goal(
        "Ship the API",
        working_dir=tmp_path,
        backend=first_backend,
        max_bounces=1,
        session_name="example",
    ) as session:
        with pytest.raises(JuvenalExecutionError):
            do(["Step one", "Step two"], checker="tester", stage_id="build-stage")

        assert session.stages["build-stage"]["status"] == "running"
        assert session.stages["build-stage"]["run_id"] == "001"
        assert [entry["instruction"] for entry in session.history] == ["Step one"]

    second_backend = MockBackend()
    second_backend.add_response(exit_code=0, output="later done")
    second_backend.add_response(exit_code=0, output="step 2 done")
    second_backend.add_response(exit_code=0, output="VERDICT: PASS")

    with goal(
        "Ship the API",
        working_dir=tmp_path,
        backend=second_backend,
        max_bounces=1,
        session_name="example",
    ) as session:
        do("Later history")
        later_summary = session.history[-1]["summary"]

        do(["Step one", "Step two"], checker="tester", stage_id="build-stage")

        assert session.stages["build-stage"]["status"] == "completed"
        assert session.stages["build-stage"]["run_id"] == "001"
        assert later_summary not in second_backend.calls[1]
        assert "Step one" in second_backend.calls[1]
        assert "Step two" in second_backend.calls[1]


def test_staged_do_rejects_reusing_stage_id_with_different_inputs(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=backend, session_name="example"):
        do("Prepare the repository", stage_id="build-stage")

        with pytest.raises(JuvenalUsageError, match="different do\\(\\) inputs"):
            do("Implement the API", stage_id="build-stage")


def test_one_shot_do_rejects_collision_with_existing_user_stage_id(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=backend, session_name="example") as session:
        do("Prepare the repository", stage_id="do-002")

        with pytest.raises(JuvenalUsageError, match="Stage 'do-002' already exists"):
            do("Implement the API")

        assert len(backend.calls) == 1
        assert session.stages["do-002"]["tasks"] == ["Prepare the repository"]
        assert session.stages["do-002"]["status"] == "completed"


def test_goal_resolves_exclude_file_via_git_in_linked_worktree(tmp_path):
    repo = tmp_path / "repo"
    _init_git_repo(repo)
    subprocess.run(["git", "config", "user.email", "juvenal@example.com"], cwd=repo, check=True)
    subprocess.run(["git", "config", "user.name", "Juvenal Tests"], cwd=repo, check=True)
    (repo / "README.md").write_text("seed\n")
    subprocess.run(["git", "add", "README.md"], cwd=repo, check=True)
    subprocess.run(["git", "commit", "-m", "init"], cwd=repo, check=True)

    worktree_dir = tmp_path / "linked-worktree"
    subprocess.run(["git", "worktree", "add", str(worktree_dir)], cwd=repo, check=True)
    exclude_file = Path(
        subprocess.run(
            ["git", "-C", str(worktree_dir), "rev-parse", "--path-format=absolute", "--git-path", "info/exclude"],
            check=True,
            capture_output=True,
            text=True,
        ).stdout.strip()
    )

    before = exclude_file.read_text() if exclude_file.exists() else ""
    before_lines = before.splitlines()

    with goal("Goal", working_dir=worktree_dir, backend=MockBackend()):
        pass

    assert (worktree_dir / ".git").is_file()
    assert exclude_file.read_text().splitlines() == before_lines + ["/.juvenal-api/"]


@pytest.mark.parametrize(
    ("working_dir_kind", "artifact_dir", "expected_entry"),
    [
        ("repo-root", None, "/.juvenal-api/"),
        ("nested", None, "/app/.juvenal-api/"),
        ("nested", Path("..") / "artifacts" / "embedded", "/artifacts/embedded/"),
        ("nested", "__outside_repo__", None),
    ],
)
def test_goal_git_exclude_entries_are_repo_relative(tmp_path, working_dir_kind, artifact_dir, expected_entry):
    repo = tmp_path / "repo"
    _init_git_repo(repo)
    app_dir = repo / "app"
    app_dir.mkdir()
    exclude_file = repo / ".git" / "info" / "exclude"
    backend = MockBackend()

    working_dir = repo if working_dir_kind == "repo-root" else app_dir
    resolved_artifact_dir = tmp_path / "outside-artifacts" if artifact_dir == "__outside_repo__" else artifact_dir
    before = exclude_file.read_text() if exclude_file.exists() else ""
    before_lines = before.splitlines()

    with goal("Goal", working_dir=working_dir, artifact_dir=resolved_artifact_dir, backend=backend):
        pass

    after = exclude_file.read_text() if exclude_file.exists() else ""
    if expected_entry is None:
        assert after == before
    else:
        assert after.splitlines() == before_lines + [expected_entry]


def test_plan_and_do_uses_session_working_dir_and_preserves_history(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="repo prepared")
    backend.add_response(exit_code=0, output="planned work complete")
    captured: dict[str, object] = {}
    planned_yaml = "name: planned\nphases:\n  - id: execute\n    prompt: 'Execute the planned work.'\n"
    wrong_yaml = "name: wrong\nphases:\n  - id: wrong\n    prompt: 'Execute the wrong workflow.'\n"

    with goal("Ship the API", working_dir=tmp_path, backend=backend) as session:
        do("Prepare the repository")
        prior_summary = session.history[0]["summary"]

        def fake_plan(**kwargs):
            captured.update(kwargs)
            workflow_path = Path(kwargs["project_dir"]) / "workflow.yaml"
            workflow_path.write_text(planned_yaml)
            returned_path = Path(kwargs["project_dir"]) / ".plan" / "planner-returned.yaml"
            returned_path.write_text(wrong_yaml)
            return PlanResult(
                success=True,
                workflow_yaml_path=str(returned_path),
                temp_dir=None,
                input_tokens=11,
                output_tokens=22,
            )

        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with patch("juvenal.api.load_workflow", wraps=load_workflow_impl) as load_workflow_mock:
                plan_and_do("Break the work into phases.")

        assert captured["project_dir"] == session.working_dir
        assert captured["backend_instance"] is backend
        assert captured["plain"] is True
        assert "Ship the API" in captured["goal"]
        assert prior_summary in captured["goal"]
        assert "Break the work into phases." in captured["goal"]
        assert load_workflow_mock.call_count == 1
        assert Path(captured["project_dir"]) / ".plan" / "planner-returned.yaml" != (tmp_path / "workflow.yaml")
        assert load_workflow_mock.call_args.args[0] == (tmp_path / "workflow.yaml").resolve()
        assert session.history[-1]["kind"] == "plan_and_do"
        assert session.history[-1]["goal_text"] == "Break the work into phases."
        workflow_archives = list(session.session_artifact_dir.glob("*-workflow.yaml"))
        planned_state_files = list(session.session_artifact_dir.glob("*-planned.json"))
        assert len(workflow_archives) == 1
        assert workflow_archives[0].read_text() == planned_yaml
        assert len(planned_state_files) == 1
        assert planned_state_files[0].exists()
        assert "Execute the planned work." in backend.calls[-1]
        assert "Execute the wrong workflow." not in backend.calls[-1]
        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["run_counter"] == 2
        assert manifest["history"][-1]["kind"] == "plan_and_do"
        assert manifest["stages"]["do-001"]["status"] == "completed"
        assert manifest["stages"]["plan-and-do-002"]["status"] == "completed"


def test_plan_and_do_load_failure_reports_workflow_yaml_path(tmp_path):
    backend = MockBackend()
    bad_yaml = "name: planned\nphases:\n  - just-a-string\n"

    with goal("Ship the API", working_dir=tmp_path, backend=backend) as session:

        def fake_plan(**kwargs):
            workflow_path = Path(kwargs["project_dir"]) / "workflow.yaml"
            workflow_path.write_text(bad_yaml)
            return PlanResult(
                success=True,
                workflow_yaml_path=str(workflow_path),
                temp_dir=None,
            )

        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with pytest.raises(JuvenalExecutionError) as exc_info:
                plan_and_do("Break the work into phases.")

        error = exc_info.value
        assert error.inspection_path == (tmp_path / "workflow.yaml").resolve()
        workflow_archives = list(session.session_artifact_dir.glob("*-workflow.yaml"))
        assert len(workflow_archives) == 1
        assert workflow_archives[0].read_text() == bad_yaml


def test_plan_and_do_planned_engine_failure_reports_state_file(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=1, output="crash")
    planned_yaml = "name: planned\nphases:\n  - id: execute\n    prompt: 'Execute the planned work.'\n"

    with goal("Ship the API", working_dir=tmp_path, backend=backend, max_bounces=1) as session:

        def fake_plan(**kwargs):
            workflow_path = Path(kwargs["project_dir"]) / "workflow.yaml"
            workflow_path.write_text(planned_yaml)
            return PlanResult(
                success=True,
                workflow_yaml_path=str(workflow_path),
                temp_dir=None,
            )

        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with pytest.raises(JuvenalExecutionError) as exc_info:
                plan_and_do("Break the work into phases.")

        error = exc_info.value
        assert error.inspection_path.parent == session.session_artifact_dir
        assert error.inspection_path.name.endswith("-planned.json")
        assert error.inspection_path.exists()
        assert session.history == []


def test_plan_and_do_planning_failure_reports_planner_state_path(tmp_path):
    backend = MockBackend()

    with goal("Ship the API", working_dir=tmp_path, backend=backend) as session:

        def fake_run(self):
            self.state.set_attempt("planner", 1)
            self.state.mark_failed("planner")
            return 1

        with patch("juvenal.engine.Engine.run", autospec=True, side_effect=fake_run):
            with pytest.raises(JuvenalExecutionError) as exc_info:
                plan_and_do("Break the work into phases.")

        error = exc_info.value
        assert error.inspection_path == (tmp_path / ".plan" / ".juvenal-state.json").resolve()
        assert error.inspection_path.exists()
        state = json.loads(error.inspection_path.read_text())
        assert state["phases"]["planner"]["attempt"] == 1
        assert state["phases"]["planner"]["status"] == "failed"
        assert session.history == []


def test_staged_plan_and_do_requires_named_session(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend()):
        with pytest.raises(JuvenalUsageError, match="named session"):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")


def test_staged_plan_and_do_rejects_second_owner_from_persisted_manifest(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        owner_path = Path(session.stages["plan-stage"]["planner_owner_path"])
        planner_state_path = Path(session.stages["plan-stage"]["planner_state_path"])
        owner_path.unlink()
        planner_state_path.unlink()

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(JuvenalUsageError, match="owner in the manifest"):
                plan_and_do("Break the work into phases.", stage_id="other-plan-stage")

        assert set(session.stages) == {"plan-stage"}


def test_staged_plan_and_do_rejects_second_owner_from_other_named_session_manifest(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        Path(session.stages["plan-stage"]["planner_owner_path"]).unlink()
        Path(session.stages["plan-stage"]["planner_state_path"]).unlink()

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="second-session") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(JuvenalUsageError, match="owner in the manifest"):
                plan_and_do("Break the work into phases.", stage_id="other-plan-stage")

        assert session.stages == {}


def test_unstaged_plan_and_do_rejects_persisted_staged_owner_from_other_session(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        Path(session.stages["plan-stage"]["planner_owner_path"]).unlink()
        Path(session.stages["plan-stage"]["planner_state_path"]).unlink()

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="second-session"):
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(
                JuvenalUsageError,
                match="Cannot mix staged and unstaged juvenal.plan_and_do\\(\\) in the same workspace",
            ):
                plan_and_do("Break the work into phases.")


def test_staged_plan_and_do_resumes_after_planner_complete_without_replanning(tmp_path):
    first_backend = MockBackend()
    first_backend.add_response(exit_code=1, output="planned run crashed")
    planned_yaml = "name: planned\nphases:\n  - id: execute\n    prompt: 'Execute the planned work.'\n"
    wrong_yaml = "name: wrong\nphases:\n  - id: wrong\n    prompt: 'Execute the wrong workflow.'\n"
    planning_calls: list[dict[str, object]] = []

    def fake_plan(**kwargs):
        planning_calls.append(kwargs)
        workflow_path = Path(kwargs["project_dir"]) / "workflow.yaml"
        workflow_path.write_text(planned_yaml)
        _write_planned_workflow_structure(Path(kwargs["project_dir"]))
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal(
        "Ship the API",
        working_dir=tmp_path,
        backend=first_backend,
        max_bounces=1,
        session_name="example",
    ) as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert len(planning_calls) == 1
        assert session.stages["plan-stage"]["status"] == "planner_complete"
        owner_path = (tmp_path / ".plan" / "staged-plan-owner.json").resolve()
        assert owner_path.exists()
        owner_path.unlink()
        (tmp_path / "workflow.yaml").write_text(wrong_yaml)

    second_backend = MockBackend()
    second_backend.add_response(exit_code=0, output="planned work complete")

    with goal(
        "Ship the API",
        working_dir=tmp_path,
        backend=second_backend,
        max_bounces=1,
        session_name="example",
    ) as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert (tmp_path / ".plan" / "staged-plan-owner.json").exists()
        assert session.stages["plan-stage"]["run_id"] == "001"
        assert session.stages["plan-stage"]["status"] == "completed"
        assert session.history[-1]["kind"] == "plan_and_do"
        assert "Execute the planned work." in second_backend.calls[0]
        assert "Execute the wrong workflow." not in second_backend.calls[0]


def test_staged_plan_and_do_rejects_planner_asset_drift_before_resume(tmp_path):
    backend = MockBackend()

    with goal("Ship the API", working_dir=tmp_path, backend=backend, session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        planner_assets_path = Path(session.stages["plan-stage"]["planner_assets_path"])
        assert session.stages["plan-stage"]["status"] == "running"
        assert planner_assets_path.exists()

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with patch(
            "juvenal.api._build_planner_assets_manifest",
            return_value={"digest": "changed", "files": [], "root": "planner"},
        ):
            with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
                with pytest.raises(JuvenalExecutionError) as exc_info:
                    plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert exc_info.value.inspection_path == planner_assets_path.resolve()


def test_planner_assets_manifest_includes_plan_validation_module():
    from juvenal.api import _build_planner_assets_manifest

    manifest = _build_planner_assets_manifest()
    manifest_paths = {entry["path"] for entry in manifest["files"]}
    expected_path = str((Path(__file__).resolve().parents[1] / "juvenal" / "plan_validation.py").resolve())

    assert expected_path in manifest_paths


def test_staged_plan_and_do_rejects_file_relative_yaml_and_rewinds_planner_state(tmp_path):
    backend = MockBackend()
    bad_yaml = (
        "name: planned\ninclude:\n  - shared.yaml\nphases:\n  - id: execute\n    prompt: 'Execute the planned work.'\n"
    )

    with goal("Ship the API", working_dir=tmp_path, backend=backend, session_name="example") as session:

        def fake_plan(**kwargs):
            workflow_path = Path(kwargs["project_dir"]) / "workflow.yaml"
            workflow_path.write_text(bad_yaml)
            _write_planned_workflow_structure(Path(kwargs["project_dir"]))
            return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with pytest.raises(JuvenalExecutionError) as exc_info:
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        error = exc_info.value
        assert error.inspection_path == (tmp_path / "workflow.yaml").resolve()
        assert session.stages["plan-stage"]["status"] == "running"
        planner_state = json.loads(Path(session.stages["plan-stage"]["planner_state_path"]).read_text())
        assert planner_state["phases"]["write-workflow"]["status"] == "pending"
        assert (
            "top-level include is not allowed"
            in planner_state["phases"]["write-workflow"]["failure_contexts"][-1]["context"]
        )


def test_goal_named_session_reuses_manifest_with_custom_artifact_root_and_exact_manifest_path(tmp_path):
    working_dir = tmp_path / "workspace"
    working_dir.mkdir()
    artifact_dir = Path("artifacts") / "embedded"
    expected_artifact_root = (working_dir / artifact_dir).resolve()
    first_backend = MockBackend()
    first_backend.add_response(exit_code=0, output="prepared")

    with goal(
        "Ship the API",
        working_dir=working_dir,
        artifact_dir=artifact_dir,
        backend=first_backend,
        session_name="example",
    ) as session:
        assert session.artifact_root == expected_artifact_root
        assert session.session_artifact_dir == (expected_artifact_root / "example").resolve()
        assert session.manifest_path == (expected_artifact_root / "example" / "session.json").resolve()

        do("Prepare the repository")

    second_backend = MockBackend()
    second_backend.add_response(exit_code=0, output="implemented")

    with goal(
        "Ship the API",
        working_dir=working_dir,
        artifact_dir=artifact_dir,
        backend=second_backend,
        session_name="example",
    ) as session:
        assert session.run_counter == 1
        assert session.manifest_path == (expected_artifact_root / "example" / "session.json").resolve()
        assert [entry["instruction"] for entry in session.history] == ["Prepare the repository"]

        do("Implement the API")

        assert [entry["instruction"] for entry in session.history] == ["Prepare the repository", "Implement the API"]


@pytest.mark.parametrize(
    "stage_id",
    ["", "Build", "has_underscore", "two--parts", "-leading", "trailing-", "with space", " spaced", "a/b", ".", ".."],
)
def test_do_rejects_invalid_stage_ids(tmp_path, stage_id):
    with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with pytest.raises(JuvenalUsageError, match="stage_id must match"):
            do("Build the feature", stage_id=stage_id)


@pytest.mark.parametrize(
    "stage_id",
    ["", "Build", "has_underscore", "two--parts", "-leading", "trailing-", "with space", " spaced", "a/b", ".", ".."],
)
def test_plan_and_do_rejects_invalid_stage_ids(tmp_path, stage_id):
    with goal("Goal", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with pytest.raises(JuvenalUsageError, match="stage_id must match"):
            plan_and_do("Break the work into phases.", stage_id=stage_id)


def test_staged_do_rejects_reusing_stage_id_with_different_checker_specs(tmp_path):
    backend = MockBackend()
    backend.add_response(exit_code=0, output="prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=backend, session_name="example"):
        do("Prepare the repository", checker="tester", stage_id="build-stage")

        with pytest.raises(JuvenalUsageError, match="different do\\(\\) inputs"):
            do("Prepare the repository", checker="pm", stage_id="build-stage")


def test_staged_do_completed_stage_is_no_op_on_resume(tmp_path):
    first_backend = MockBackend()
    first_backend.add_response(exit_code=0, output="prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=first_backend, session_name="example") as session:
        do("Prepare the repository", stage_id="build-stage")
        first_history = list(session.history)

    second_backend = MockBackend()
    with goal("Ship the API", working_dir=tmp_path, backend=second_backend, session_name="example") as session:
        do("Prepare the repository", stage_id="build-stage")

        assert second_backend.calls == []
        assert second_backend.resume_calls == []
        assert session.history == first_history
        assert session.stages["build-stage"]["status"] == "completed"


def test_staged_do_zero_attempt_resume_uses_precreated_state_file(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api.Engine.run", side_effect=RuntimeError("crashed before first attempt")):
            with pytest.raises(JuvenalExecutionError, match="Embedded engine failed"):
                do("Prepare the repository", stage_id="build-stage")

        state_file = Path(session.stages["build-stage"]["state_file"])
        assert state_file.exists()
        assert PipelineState.load(state_file).phases == {}
        assert session.stages["build-stage"]["status"] == "running"
        assert session.stages["build-stage"]["run_id"] == "001"

    resume_backend = MockBackend()
    resume_backend.add_response(exit_code=0, output="prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=resume_backend, session_name="example") as session:
        do("Prepare the repository", stage_id="build-stage")

        assert session.stages["build-stage"]["status"] == "completed"
        assert session.stages["build-stage"]["run_id"] == "001"
        assert resume_backend.resume_calls == []
        assert len(resume_backend.calls) == 1


@pytest.mark.parametrize("state_mode", ["missing", "corrupt"])
def test_staged_do_missing_or_corrupt_state_file_reports_recorded_path(tmp_path, state_mode):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api.Engine.run", side_effect=RuntimeError("crashed before first attempt")):
            with pytest.raises(JuvenalExecutionError):
                do("Prepare the repository", stage_id="build-stage")

        state_file = Path(session.stages["build-stage"]["state_file"])
        if state_mode == "missing":
            state_file.unlink()
        else:
            state_file.write_text("{not json", encoding="utf-8")

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with pytest.raises(
            JuvenalExecutionError,
            match="Recorded do\\(\\) state file is missing or unreadable",
        ) as exc_info:
            do("Prepare the repository", stage_id="build-stage")

    assert exc_info.value.inspection_path == state_file.resolve()


def test_staged_plan_and_do_rejects_reusing_stage_id_with_different_goal_text(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        with pytest.raises(JuvenalUsageError, match="different plan_and_do\\(\\) goal"):
            plan_and_do("Break the work into smaller phases.", stage_id="plan-stage")


def test_staged_plan_and_do_rejects_orphan_owner_file_without_manifest_stage(tmp_path):
    owner_path = (tmp_path / ".plan" / "staged-plan-owner.json").resolve()
    owner_path.parent.mkdir(parents=True)
    owner_path.write_text(
        json.dumps(
            {
                "session_id": "someone-else",
                "session_name": "someone-else",
                "stage_id": "other-plan-stage",
                "run_id": "999",
            }
        ),
        encoding="utf-8",
    )

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with pytest.raises(JuvenalUsageError, match="Workspace already has a staged juvenal.plan_and_do\\(\\) owner"):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")


def test_staged_plan_and_do_rejects_orphan_planner_state_without_owner(tmp_path):
    planner_state_path = (tmp_path / ".plan" / ".juvenal-state.json").resolve()
    planner_state_path.parent.mkdir(parents=True)
    PipelineState(state_file=planner_state_path).save()

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with pytest.raises(JuvenalUsageError, match="planner state already exists without a staged owner"):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")


def test_staged_plan_and_do_rejects_second_active_stage_while_owner_file_exists(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(
                JuvenalUsageError,
                match="Workspace already has a staged juvenal.plan_and_do\\(\\) owner",
            ):
                plan_and_do("Break the work into smaller phases.", stage_id="second-plan-stage")


def test_staged_plan_and_do_completed_stage_is_no_op_on_resume(tmp_path):
    def fake_plan(**kwargs):
        workflow_path = _write_planned_workflow(Path(kwargs["project_dir"]), "name: planned\nphases: []\n", phases=[])
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        first_history = list(session.history)

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert session.history == first_history
        assert session.stages["plan-stage"]["status"] == "completed"


def test_staged_plan_and_do_zero_attempt_planner_resume_uses_stored_goal_snapshot(tmp_path):
    prepare_backend = MockBackend()
    prepare_backend.add_response(exit_code=0, output="prepared")

    with goal("Ship the API", working_dir=tmp_path, backend=prepare_backend, session_name="example") as session:
        do("Prepare the repository")
        initial_summary = session.history[-1]["summary"]

        with patch(
            "juvenal.api._plan_workflow_internal",
            side_effect=RuntimeError("planner crashed before first attempt"),
        ):
            with pytest.raises(JuvenalExecutionError, match="Planning failed"):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        planner_state_path = Path(session.stages["plan-stage"]["planner_state_path"])
        assert PipelineState.load(planner_state_path).phases == {}

    resume_backend = MockBackend()
    resume_backend.add_response(exit_code=0, output="later history")
    captured: dict[str, object] = {}

    with goal("Ship the API", working_dir=tmp_path, backend=resume_backend, session_name="example") as session:
        do("Later history")
        later_summary = session.history[-1]["summary"]

        def fake_plan(**kwargs):
            captured.update(kwargs)
            workflow_path = _write_planned_workflow(
                Path(kwargs["project_dir"]),
                "name: planned\nphases: []\n",
                phases=[],
            )
            return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert captured["resume"] is True
        assert initial_summary in captured["goal"]
        assert later_summary not in captured["goal"]
        assert session.stages["plan-stage"]["status"] == "completed"
        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["stages"]["plan-stage"]["status"] == "completed"
        assert manifest["history"][-1]["kind"] == "plan_and_do"


@pytest.mark.parametrize(
    ("asset_mode", "message"),
    [
        ("missing", "Planner assets manifest is missing"),
        ("corrupt", "Planner assets manifest is unreadable"),
    ],
)
def test_staged_plan_and_do_missing_or_corrupt_planner_assets_manifest_fails_before_resume(
    tmp_path,
    asset_mode,
    message,
):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        planner_assets_path = Path(session.stages["plan-stage"]["planner_assets_path"])
        if asset_mode == "missing":
            planner_assets_path.unlink()
        else:
            planner_assets_path.write_text("{not json", encoding="utf-8")

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(JuvenalExecutionError, match=message) as exc_info:
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

    assert exc_info.value.inspection_path == planner_assets_path.resolve()


def test_staged_plan_and_do_missing_planner_state_file_reports_recorded_path(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        planner_state_path = Path(session.stages["plan-stage"]["planner_state_path"])
        planner_state_path.unlink()

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with pytest.raises(
            JuvenalExecutionError,
            match="Recorded planner state file is missing or unreadable",
        ) as exc_info:
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

    assert exc_info.value.inspection_path == planner_state_path.resolve()


def test_staged_plan_and_do_recreates_missing_owner_file_during_running_stage_resume(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        owner_path = Path(session.stages["plan-stage"]["planner_owner_path"])
        owner_path.unlink()

    def fake_plan(**kwargs):
        workflow_path = _write_planned_workflow(Path(kwargs["project_dir"]), "name: planned\nphases: []\n", phases=[])
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert owner_path.exists()
        assert session.stages["plan-stage"]["status"] == "completed"


def test_staged_plan_and_do_rejects_owner_file_mismatch_on_resume(tmp_path):
    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch(
            "juvenal.api._plan_workflow_internal",
            return_value=PlanResult(success=False, error="planner failed", temp_dir=None),
        ):
            with pytest.raises(JuvenalExecutionError):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        owner_path = Path(session.stages["plan-stage"]["planner_owner_path"])
        owner_path.write_text(
            json.dumps(
                {
                    "session_id": "example",
                    "session_name": "example",
                    "stage_id": "other-plan-stage",
                    "run_id": "001",
                }
            ),
            encoding="utf-8",
        )

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(JuvenalUsageError, match="different staged juvenal.plan_and_do\\(\\) owner"):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")


@pytest.mark.parametrize(
    ("artifact_key", "expected_message"),
    [
        ("workflow_archive_path", "Recorded archived workflow is missing or unreadable"),
        ("planned_state_path", "Recorded planned execution state file is missing or unreadable"),
    ],
)
def test_staged_plan_and_do_rejects_deleted_planner_complete_artifacts(tmp_path, artifact_key, expected_message):
    planned_yaml = "name: planned\nphases:\n  - id: execute\n    prompt: 'Execute the planned work.'\n"

    def fake_plan(**kwargs):
        workflow_path = _write_planned_workflow(Path(kwargs["project_dir"]), planned_yaml)
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with patch("juvenal.api.Engine.run", side_effect=RuntimeError("planned run crashed before phase 1")):
                with pytest.raises(JuvenalExecutionError):
                    plan_and_do("Break the work into phases.", stage_id="plan-stage")

        artifact_path = Path(session.stages["plan-stage"][artifact_key])
        artifact_path.unlink()
        assert session.stages["plan-stage"]["status"] == "planner_complete"

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example"):
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            with pytest.raises(JuvenalExecutionError, match=expected_message) as exc_info:
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

    assert exc_info.value.inspection_path == artifact_path.resolve()


def test_staged_plan_and_do_planner_complete_zero_phase_resume_completes_and_persists_history(tmp_path):
    def fake_plan(**kwargs):
        workflow_path = _write_planned_workflow(Path(kwargs["project_dir"]), "name: planned\nphases: []\n", phases=[])
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            with patch("juvenal.api.Engine.run", side_effect=RuntimeError("planned run crashed before phase 1")):
                with pytest.raises(JuvenalExecutionError, match="Planned engine failed"):
                    plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert session.stages["plan-stage"]["status"] == "planner_complete"

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=AssertionError("planner should not rerun")):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert session.stages["plan-stage"]["status"] == "completed"
        assert session.history[-1]["kind"] == "plan_and_do"
        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["stages"]["plan-stage"]["status"] == "completed"
        assert manifest["history"][-1]["kind"] == "plan_and_do"


def test_staged_plan_and_do_completion_checkpoint_persists_status_and_history_together(tmp_path):
    snapshots: list[tuple[str | None, tuple[str, ...]]] = []
    real_save = api_module._save_session_manifest

    def tracking_save(session):
        stage_record = session.stages.get("plan-stage")
        stage_status = stage_record.get("status") if isinstance(stage_record, dict) else None
        history_kinds = tuple(entry.get("kind", "") for entry in session.history)
        snapshots.append((stage_status, history_kinds))
        return real_save(session)

    def fake_plan(**kwargs):
        workflow_path = _write_planned_workflow(Path(kwargs["project_dir"]), "name: planned\nphases: []\n", phases=[])
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal("Ship the API", working_dir=tmp_path, backend=MockBackend(), session_name="example") as session:
        with patch("juvenal.api._save_session_manifest", side_effect=tracking_save):
            with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
                plan_and_do("Break the work into phases.", stage_id="plan-stage")

        completed_indices = [index for index, (status, _history_kinds) in enumerate(snapshots) if status == "completed"]
        history_indices = [
            index for index, (_status, history_kinds) in enumerate(snapshots) if history_kinds == ("plan_and_do",)
        ]

        assert completed_indices == [len(snapshots) - 1]
        assert history_indices == [len(snapshots) - 1]
        assert snapshots[-2] == ("planner_complete", ())
        assert snapshots[-1] == ("completed", ("plan_and_do",))
        manifest = json.loads(session.manifest_path.read_text())
        assert manifest["stages"]["plan-stage"]["status"] == "completed"
        assert manifest["history"][-1]["kind"] == "plan_and_do"


def test_staged_plan_and_do_ignores_anonymous_plan_and_do_sessions_in_other_working_dirs(tmp_path):
    artifact_root = tmp_path / "shared-artifacts"
    first_workspace = tmp_path / "first"
    second_workspace = tmp_path / "second"
    first_workspace.mkdir()
    second_workspace.mkdir()

    def fake_plan(**kwargs):
        workflow_path = _write_planned_workflow(Path(kwargs["project_dir"]), "name: planned\nphases: []\n", phases=[])
        return PlanResult(success=True, workflow_yaml_path=str(workflow_path), temp_dir=None)

    with goal("Workspace one", working_dir=first_workspace, artifact_dir=artifact_root, backend=MockBackend()):
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            plan_and_do("Anonymous planning run.")

    with goal(
        "Workspace two",
        working_dir=second_workspace,
        artifact_dir=artifact_root,
        backend=MockBackend(),
        session_name="example",
    ) as session:
        with patch("juvenal.api._plan_workflow_internal", side_effect=fake_plan):
            plan_and_do("Break the work into phases.", stage_id="plan-stage")

        assert session.stages["plan-stage"]["status"] == "completed"



"""Public embedded API for Juvenal."""

from __future__ import annotations

import contextvars
import hashlib
import json
import os
import re
import subprocess
from collections.abc import Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import count
from pathlib import Path
from typing import Any, Iterator

import yaml

from juvenal.backends import Backend, create_backend
from juvenal.engine import Engine, _plan_workflow_internal
from juvenal.plan_validation import validate_planned_workflow
from juvenal.state import PipelineState
from juvenal.workflow import Phase, Workflow, inject_checkers, load_workflow, parse_checker_string, validate_workflow


class JuvenalUsageError(Exception):
    """Raised for deterministic API misuse."""


class JuvenalExecutionError(Exception):
    """Raised when planning, loading, validation, or execution fails."""

    def __init__(self, message: str, *, run_id: str, inspection_path: str | Path):
        self.run_id = run_id
        self.inspection_path = Path(inspection_path).resolve()
        super().__init__(f"{message} [run_id={run_id}, inspect={self.inspection_path}]")


@dataclass
class GoalSession:
    """Embedded API session state."""

    goal_text: str
    working_dir: Path
    backend_name: str
    backend_instance: Backend
    max_bounces: int
    plain: bool
    serialize: bool
    clear_context_on_bounce: bool
    artifact_root: Path
    session_id: str
    session_name: str | None
    session_artifact_dir: Path
    manifest_path: Path
    run_counter: int = 0
    history: list[dict[str, Any]] = field(default_factory=list)
    stages: dict[str, dict[str, Any]] = field(default_factory=dict)

    @property
    def session_key(self) -> str:
        return self.session_id


@dataclass(frozen=True)
class _GitContext:
    repo_root: Path
    exclude_file: Path


_ACTIVE_SESSION: contextvars.ContextVar[GoalSession | None] = contextvars.ContextVar(
    "juvenal_active_session",
    default=None,
)
_DO_HISTORY_LIMIT = 10
_DO_SUMMARY_LIMIT = 200
_SESSION_MANIFEST_FILENAME = "session.json"
_SESSION_MANIFEST_SCHEMA_VERSION = 1
_STAGED_PLAN_OWNER_FILENAME = "staged-plan-owner.json"
_STAGED_PLAN_WRITE_WORKFLOW_PHASE_ID = "write-workflow"
_IDENTIFIER_RE = re.compile(r"^[a-z0-9]+(?:-[a-z0-9]+)*$")
_RESERVED_SESSION_NAME_RE = re.compile(r"^session-[0-9]+$")
_SESSION_IDENTITY_FIELDS = (
    "goal_text",
    "working_dir",
    "backend_name",
    "max_bounces",
    "serialize",
    "clear_context_on_bounce",
)


def _resolve_working_dir(working_dir: str | Path) -> Path:
    resolved = Path(working_dir).expanduser().resolve()
    if not resolved.exists():
        raise JuvenalUsageError(f"Working directory does not exist: {resolved}")
    if not resolved.is_dir():
        raise JuvenalUsageError(f"Working directory is not a directory: {resolved}")
    return resolved


def _resolve_backend(backend: str | Backend) -> tuple[str, Backend]:
    if isinstance(backend, str):
        try:
            return backend, create_backend(backend)
        except ValueError as exc:
            raise JuvenalUsageError(str(exc)) from exc
    if isinstance(backend, Backend):
        return backend.name(), backend
    raise JuvenalUsageError(f"Unsupported backend value: {backend!r}")


def _resolve_artifact_root(working_dir: Path, artifact_dir: str | Path | None) -> Path:
    if artifact_dir is None:
        artifact_root = working_dir / ".juvenal-api"
    else:
        artifact_root = Path(artifact_dir)
        if not artifact_root.is_absolute():
            artifact_root = working_dir / artifact_root
    artifact_root = artifact_root.expanduser().resolve()
    try:
        artifact_root.mkdir(parents=True, exist_ok=True)
    except FileExistsError as exc:
        raise JuvenalUsageError(f"Artifact root exists and is not a directory: {artifact_root}") from exc
    return artifact_root


def _validate_identifier(value: str, *, field_name: str) -> str:
    if not isinstance(value, str):
        raise JuvenalUsageError(f"{field_name} must be a string, got {type(value).__name__}")
    if not _IDENTIFIER_RE.fullmatch(value):
        raise JuvenalUsageError(f"{field_name} must match ^[a-z0-9]+(?:-[a-z0-9]+)*$: {value!r}")
    return value


def _validate_session_name(session_name: str) -> str:
    validated = _validate_identifier(session_name, field_name="session_name")
    if _is_reserved_session_name(validated):
        raise JuvenalUsageError(f"session_name {session_name!r} is reserved for anonymous sessions")
    return validated


def _validate_stage_id(stage_id: str) -> str:
    return _validate_identifier(stage_id, field_name="stage_id")


def _is_reserved_session_name(session_name: str) -> bool:
    return _RESERVED_SESSION_NAME_RE.fullmatch(session_name) is not None


def _session_manifest_path(session_dir: Path) -> Path:
    return (session_dir / _SESSION_MANIFEST_FILENAME).resolve()


def _resolve_anonymous_session_dir(artifact_root: Path) -> tuple[str, Path]:
    session_id, session_dir = _allocate_session_dir(artifact_root)
    return session_id, session_dir.resolve()


def _resolve_named_session_dir(artifact_root: Path, session_name: str) -> tuple[str, Path, bool]:
    validated_name = _validate_session_name(session_name)
    session_dir = (artifact_root / validated_name).resolve()
    existed = session_dir.exists()

    if existed and not session_dir.is_dir():
        raise JuvenalUsageError(f"Session artifact path exists and is not a directory: {session_dir}")
    if not existed:
        session_dir.mkdir(exist_ok=False)
    return validated_name, session_dir, existed


def _session_identity_values(
    *,
    goal_text: str,
    working_dir: Path,
    backend_name: str,
    max_bounces: int,
    serialize: bool,
    clear_context_on_bounce: bool,
) -> dict[str, Any]:
    return {
        "goal_text": goal_text,
        "working_dir": str(working_dir),
        "backend_name": backend_name,
        "max_bounces": max_bounces,
        "serialize": serialize,
        "clear_context_on_bounce": clear_context_on_bounce,
    }


def _session_manifest_payload(session: GoalSession) -> dict[str, Any]:
    return {
        "schema_version": _SESSION_MANIFEST_SCHEMA_VERSION,
        "session_id": session.session_id,
        "session_name": session.session_name,
        **_session_identity_values(
            goal_text=session.goal_text,
            working_dir=session.working_dir,
            backend_name=session.backend_name,
            max_bounces=session.max_bounces,
            serialize=session.serialize,
            clear_context_on_bounce=session.clear_context_on_bounce,
        ),
        "run_counter": session.run_counter,
        "history": session.history,
        "stages": session.stages,
    }


def _invalid_session_manifest(manifest_path: Path, detail: str) -> JuvenalUsageError:
    return JuvenalUsageError(f"Invalid session manifest at {manifest_path}: {detail}")


def _write_json_atomic(path: Path, payload: dict[str, Any]) -> None:
    serialized = json.dumps(payload, indent=2, sort_keys=True) + "\n"
    path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = path.with_name(f"{path.name}.tmp")
    with open(tmp_path, "w", encoding="utf-8") as f:
        f.write(serialized)
        f.flush()
        os.fsync(f.fileno())
    os.replace(tmp_path, path)


def _save_session_manifest(session: GoalSession) -> None:
    _write_json_atomic(session.manifest_path, _session_manifest_payload(session))


def _load_session_manifest(manifest_path: Path) -> dict[str, Any]:
    try:
        raw = json.loads(manifest_path.read_text(encoding="utf-8"))
    except FileNotFoundError as exc:
        raise _invalid_session_manifest(manifest_path, "missing session.json") from exc
    except (OSError, json.JSONDecodeError) as exc:
        raise _invalid_session_manifest(manifest_path, str(exc)) from exc

    if not isinstance(raw, dict):
        raise _invalid_session_manifest(manifest_path, "manifest root must be a JSON object")
    if raw.get("schema_version") != _SESSION_MANIFEST_SCHEMA_VERSION:
        raise _invalid_session_manifest(
            manifest_path,
            f"unsupported schema_version {raw.get('schema_version')!r}",
        )

    session_id = raw.get("session_id")
    try:
        _validate_identifier(session_id, field_name="session manifest session_id")
    except JuvenalUsageError as exc:
        raise _invalid_session_manifest(manifest_path, str(exc)) from exc

    session_name = raw.get("session_name")
    if session_name is not None:
        try:
            _validate_session_name(session_name)
        except JuvenalUsageError as exc:
            raise _invalid_session_manifest(manifest_path, str(exc)) from exc
        if session_name != session_id:
            raise _invalid_session_manifest(
                manifest_path,
                f"session_name {session_name!r} must match session_id {session_id!r}",
            )
    elif not _is_reserved_session_name(session_id):
        raise _invalid_session_manifest(manifest_path, "session_name is required for named sessions")

    if session_id != manifest_path.parent.name:
        raise _invalid_session_manifest(
            manifest_path,
            f"session_id {session_id!r} does not match directory {manifest_path.parent.name!r}",
        )

    goal_text = raw.get("goal_text")
    if not isinstance(goal_text, str):
        raise _invalid_session_manifest(manifest_path, "goal_text must be a string")

    working_dir = raw.get("working_dir")
    if not isinstance(working_dir, str):
        raise _invalid_session_manifest(manifest_path, "working_dir must be a string")

    backend_name = raw.get("backend_name")
    if not isinstance(backend_name, str):
        raise _invalid_session_manifest(manifest_path, "backend_name must be a string")

    max_bounces = raw.get("max_bounces")
    if type(max_bounces) is not int:
        raise _invalid_session_manifest(manifest_path, "max_bounces must be an int")

    serialize = raw.get("serialize")
    if not isinstance(serialize, bool):
        raise _invalid_session_manifest(manifest_path, "serialize must be a bool")

    clear_context_on_bounce = raw.get("clear_context_on_bounce")
    if not isinstance(clear_context_on_bounce, bool):
        raise _invalid_session_manifest(manifest_path, "clear_context_on_bounce must be a bool")

    run_counter = raw.get("run_counter")
    if type(run_counter) is not int or run_counter < 0:
        raise _invalid_session_manifest(manifest_path, "run_counter must be a non-negative int")

    history = raw.get("history")
    if not isinstance(history, list):
        raise _invalid_session_manifest(manifest_path, "history must be a list")
    normalized_history: list[dict[str, Any]] = []
    for index, entry in enumerate(history, start=1):
        if not isinstance(entry, dict):
            raise _invalid_session_manifest(manifest_path, f"history entry {index} must be an object")
        normalized_history.append(dict(entry))

    stages_raw = raw.get("stages")
    if not isinstance(stages_raw, dict):
        raise _invalid_session_manifest(manifest_path, "stages must be an object")

    stages: dict[str, dict[str, Any]] = {}
    for stage_id, stage_data in stages_raw.items():
        if not isinstance(stage_id, str):
            raise _invalid_session_manifest(manifest_path, "stage ids must be strings")
        try:
            _validate_stage_id(stage_id)
        except JuvenalUsageError as exc:
            raise _invalid_session_manifest(manifest_path, str(exc)) from exc
        if not isinstance(stage_data, dict):
            raise _invalid_session_manifest(manifest_path, f"stage {stage_id!r} must map to an object")
        stages[stage_id] = dict(stage_data)

    return {
        "schema_version": _SESSION_MANIFEST_SCHEMA_VERSION,
        "session_id": session_id,
        "session_name": session_name,
        "goal_text": goal_text,
        "working_dir": working_dir,
        "backend_name": backend_name,
        "max_bounces": max_bounces,
        "serialize": serialize,
        "clear_context_on_bounce": clear_context_on_bounce,
        "run_counter": run_counter,
        "history": normalized_history,
        "stages": stages,
    }


def _assert_session_identity_matches(
    *,
    session_name: str,
    manifest_path: Path,
    manifest: dict[str, Any],
    goal_text: str,
    working_dir: Path,
    backend_name: str,
    max_bounces: int,
    serialize: bool,
    clear_context_on_bounce: bool,
) -> None:
    expected = _session_identity_values(
        goal_text=goal_text,
        working_dir=working_dir,
        backend_name=backend_name,
        max_bounces=max_bounces,
        serialize=serialize,
        clear_context_on_bounce=clear_context_on_bounce,
    )
    mismatches = [
        field_name for field_name in _SESSION_IDENTITY_FIELDS if manifest.get(field_name) != expected[field_name]
    ]
    if mismatches:
        mismatch_list = ", ".join(mismatches)
        raise JuvenalUsageError(
            f"Named session {session_name!r} does not match {manifest_path}: identity mismatch in {mismatch_list}"
        )


def _build_goal_session(
    *,
    goal_text: str,
    working_dir: Path,
    backend_name: str,
    backend_instance: Backend,
    max_bounces: int,
    plain: bool,
    serialize: bool,
    clear_context_on_bounce: bool,
    artifact_root: Path,
    session_id: str,
    session_name: str | None,
    session_artifact_dir: Path,
    manifest_path: Path,
    run_counter: int = 0,
    history: list[dict[str, Any]] | None = None,
    stages: dict[str, dict[str, Any]] | None = None,
) -> GoalSession:
    return GoalSession(
        goal_text=goal_text,
        working_dir=working_dir,
        backend_name=backend_name,
        backend_instance=backend_instance,
        max_bounces=max_bounces,
        plain=plain,
        serialize=serialize,
        clear_context_on_bounce=clear_context_on_bounce,
        artifact_root=artifact_root,
        session_id=session_id,
        session_name=session_name,
        session_artifact_dir=session_artifact_dir,
        manifest_path=manifest_path,
        run_counter=run_counter,
        history=list(history or []),
        stages=dict(stages or {}),
    )


def _run_git(working_dir: Path, *args: str) -> str | None:
    try:
        result = subprocess.run(
            ["git", "-C", str(working_dir), *args],
            capture_output=True,
            text=True,
            timeout=10,
        )
    except (FileNotFoundError, subprocess.SubprocessError):
        return None
    if result.returncode != 0:
        return None
    output = result.stdout.strip()
    return output or None


def _resolve_git_context(working_dir: Path) -> _GitContext | None:
    repo_root_output = _run_git(working_dir, "rev-parse", "--show-toplevel")
    if repo_root_output is None:
        return None

    repo_root = Path(repo_root_output).resolve()
    exclude_output = _run_git(working_dir, "rev-parse", "--path-format=absolute", "--git-path", "info/exclude")
    exclude_file: Path | None = None

    if exclude_output:
        candidate = Path(exclude_output)
        if candidate.is_absolute():
            exclude_file = candidate.resolve()
        else:
            exclude_file = (working_dir / candidate).resolve()

    if exclude_file is None:
        raw_output = _run_git(working_dir, "rev-parse", "--git-path", "info/exclude")
        if raw_output is None:
            return None
        exclude_file = (working_dir / raw_output).resolve()

    return _GitContext(repo_root=repo_root, exclude_file=exclude_file)


def _git_ignore_entry(repo_root: Path, path: Path) -> str:
    rel_path = path.resolve().relative_to(repo_root).as_posix()
    if not rel_path:
        return "/" if not path.is_dir() else "/"
    if path.is_dir():
        return f"/{rel_path}/"
    return f"/{rel_path}"


def _ensure_git_excluded(git_context: _GitContext, path: Path) -> None:
    try:
        path.resolve().relative_to(git_context.repo_root)
    except ValueError:
        return

    entry = _git_ignore_entry(git_context.repo_root, path)
    git_context.exclude_file.parent.mkdir(parents=True, exist_ok=True)
    if git_context.exclude_file.exists():
        existing_text = git_context.exclude_file.read_text()
        existing_lines = existing_text.splitlines()
    else:
        existing_text = ""
        existing_lines = []

    if entry in existing_lines:
        return

    prefix = ""
    if existing_text and not existing_text.endswith("\n"):
        prefix = "\n"
    with git_context.exclude_file.open("a") as f:
        f.write(f"{prefix}{entry}\n")


def _allocate_session_dir(artifact_root: Path) -> tuple[str, Path]:
    for index in count(1):
        session_id = f"session-{index:03d}"
        session_dir = artifact_root / session_id
        try:
            session_dir.mkdir(exist_ok=False)
        except FileExistsError:
            continue
        return session_id, session_dir
    raise AssertionError("unreachable")


def _require_active_session(api_name: str) -> GoalSession:
    session = _ACTIVE_SESSION.get()
    if session is None:
        raise JuvenalUsageError(f"juvenal.{api_name}() requires an active juvenal.goal(...) session")
    return session


def _allocate_run_id(session: GoalSession) -> str:
    session.run_counter += 1
    return f"{session.run_counter:03d}"


def _stage_id_for_run(stage_prefix: str, run_id: str) -> str:
    _validate_stage_id(stage_prefix)
    return _validate_stage_id(f"{stage_prefix}-{run_id}")


def _create_stage_record(session: GoalSession, *, stage_id: str, stage_record: dict[str, Any]) -> None:
    validated_stage_id = _validate_stage_id(stage_id)
    if validated_stage_id in session.stages:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} already exists")
    session.stages[validated_stage_id] = dict(stage_record)
    _save_session_manifest(session)


def _persist_session_checkpoint(
    session: GoalSession,
    *,
    stage_id: str | None = None,
    stage_updates: dict[str, Any] | None = None,
    history_entries: Sequence[dict[str, Any]] | None = None,
) -> None:
    if stage_id is not None:
        validated_stage_id = _validate_stage_id(stage_id)
        if validated_stage_id not in session.stages:
            raise AssertionError(f"Missing stage record for checkpoint: {validated_stage_id}")
        if stage_updates:
            session.stages[validated_stage_id].update(stage_updates)

    if history_entries:
        session.history.extend(history_entries)

    _save_session_manifest(session)


def _initialize_empty_pipeline_state(state_file: Path) -> None:
    PipelineState(state_file=state_file.resolve()).save()


def _recorded_stage_path(stage_id: str, stage_record: dict[str, Any], key: str) -> Path:
    raw_path = stage_record.get(key)
    if not isinstance(raw_path, str) or not raw_path:
        raise JuvenalUsageError(f"Stage {stage_id!r} is missing recorded path {key!r}")
    return Path(raw_path).expanduser().resolve()


def _raise_missing_stage_artifact(run_id: str, path: Path, detail: str) -> JuvenalExecutionError:
    return JuvenalExecutionError(
        detail,
        run_id=_run_label(run_id),
        inspection_path=path,
    )


def _ensure_required_file_exists(run_id: str, path: Path, *, detail: str) -> None:
    try:
        if not path.exists() or not path.is_file():
            raise _raise_missing_stage_artifact(run_id, path, detail)
        path.read_bytes()
    except JuvenalExecutionError:
        raise
    except OSError as exc:
        raise _raise_missing_stage_artifact(run_id, path, f"{detail}: {exc}") from exc


def _ensure_required_state_file(run_id: str, path: Path, *, detail: str) -> None:
    if not path.exists() or not path.is_file():
        raise _raise_missing_stage_artifact(run_id, path, detail)
    try:
        PipelineState.load(path)
    except Exception as exc:
        raise _raise_missing_stage_artifact(run_id, path, f"{detail}: {exc}") from exc


def _run_label(run_id: str) -> str:
    return f"run-{run_id}"


def _normalize_do_tasks(task_or_tasks: str | Sequence[str]) -> list[str]:
    if isinstance(task_or_tasks, str):
        raw_tasks = [task_or_tasks]
    elif isinstance(task_or_tasks, Sequence):
        raw_tasks = list(task_or_tasks)
        if not raw_tasks:
            raise JuvenalUsageError("juvenal.do() requires at least one task")
    else:
        raise JuvenalUsageError("juvenal.do() expects a task string or a sequence of task strings")

    tasks: list[str] = []
    for index, raw_task in enumerate(raw_tasks, start=1):
        if not isinstance(raw_task, str):
            raise JuvenalUsageError(f"juvenal.do() task {index} must be a string, got {type(raw_task).__name__}")
        task = raw_task.strip()
        if not task:
            raise JuvenalUsageError(f"juvenal.do() task {index} is empty after stripping whitespace")
        tasks.append(task)
    return tasks


def _normalize_checker_specs(checker: str | None, checkers: Sequence[str] | None) -> list[str]:
    if checker is not None and checkers is not None:
        raise JuvenalUsageError("juvenal.do() accepts either checker= or checkers=, not both")
    if checker is not None and not isinstance(checker, str):
        raise JuvenalUsageError(f"juvenal.do() checker= must be a string, got {type(checker).__name__}")
    if checkers is None:
        return [checker] if checker is not None else []
    if isinstance(checkers, str):
        raise JuvenalUsageError("juvenal.do() checkers= must be a sequence of checker specs, not a bare string")
    if not isinstance(checkers, Sequence):
        raise JuvenalUsageError("juvenal.do() checkers= must be a sequence of checker specs")

    normalized = list(checkers)
    for index, spec in enumerate(normalized, start=1):
        if not isinstance(spec, str):
            raise JuvenalUsageError(f"juvenal.do() checker spec {index} must be a string, got {type(spec).__name__}")
    return normalized


def _normalize_history_summary(text: str, limit: int = _DO_SUMMARY_LIMIT) -> str:
    normalized = " ".join(text.split())
    if len(normalized) <= limit:
        return normalized
    return normalized[: limit - 3].rstrip() + "..."


def _build_history_summary(phase_id: str, instruction: str) -> str:
    phase_label = phase_id.replace("-", " ")
    return _normalize_history_summary(f"{phase_label}: {instruction}")


def _recent_history_summaries(session: GoalSession) -> list[str]:
    summaries = [entry["summary"] for entry in session.history if isinstance(entry.get("summary"), str)]
    return summaries[-_DO_HISTORY_LIMIT:]


def _format_recent_history_block(recent_history_summaries: Sequence[str]) -> str:
    if recent_history_summaries:
        return "\n".join(f"- {summary}" for summary in recent_history_summaries)
    return "- None."


def _build_do_prompt(
    *,
    session: GoalSession,
    recent_history_summaries: Sequence[str],
    completed_steps: Sequence[str],
    instruction: str,
) -> str:
    recent_history_block = _format_recent_history_block(recent_history_summaries)

    if completed_steps:
        completed_steps_block = "\n".join(
            f"- Step {index}: {step}" for index, step in enumerate(completed_steps, start=1)
        )
    else:
        completed_steps_block = "- None."

    return (
        "You are executing one validated Juvenal embedded implementation step.\n\n"
        "Session Goal:\n"
        f"{session.goal_text}\n\n"
        "Absolute Working Directory:\n"
        f"{session.working_dir}\n\n"
        "Recent Successful Session History Summaries:\n"
        f"{recent_history_block}\n\n"
        "Earlier steps in this do() call:\n"
        f"{completed_steps_block}\n\n"
        "Current Instruction:\n"
        f"{instruction}\n"
    )


def _normalize_plan_goal_text(goal_text: str) -> tuple[str, str]:
    if not isinstance(goal_text, str):
        raise JuvenalUsageError(f"juvenal.plan_and_do() goal_text must be a string, got {type(goal_text).__name__}")

    stripped_goal_text = goal_text.strip()
    if not stripped_goal_text:
        raise JuvenalUsageError("juvenal.plan_and_do() goal_text is empty after stripping whitespace")

    normalized_goal_text = " ".join(stripped_goal_text.split())
    return stripped_goal_text, normalized_goal_text


def _build_plan_and_do_goal(
    *,
    session: GoalSession,
    recent_history_summaries: Sequence[str],
    goal_text: str,
) -> str:
    recent_history_block = _format_recent_history_block(recent_history_summaries)
    return (
        "Session Goal:\n"
        f"{session.goal_text}\n\n"
        "Recent Successful Session History Summaries:\n"
        f"{recent_history_block}\n\n"
        "Local Planning Request:\n"
        f"{goal_text}\n"
    )


def _phase_completed_with_children(workflow: Workflow, engine: Engine, phase_id: str) -> bool:
    phase_state = engine.state.phases.get(phase_id)
    if phase_state is None or phase_state.status != "completed":
        return False

    child_prefix = f"{phase_id}~"
    child_ids = [phase.id for phase in workflow.phases if phase.id.startswith(child_prefix)]
    for child_id in child_ids:
        child_state = engine.state.phases.get(child_id)
        if child_state is None or child_state.status != "completed":
            return False
    return True


def _build_do_stored_steps(session: GoalSession, *, run_id: str, tasks: Sequence[str]) -> list[dict[str, str]]:
    recent_history_summaries = _recent_history_summaries(session)
    stored_steps: list[dict[str, str]] = []

    for index, instruction in enumerate(tasks, start=1):
        phase_id = f"do-{run_id}-step-{index}"
        prompt = _build_do_prompt(
            session=session,
            recent_history_summaries=recent_history_summaries,
            completed_steps=tasks[: index - 1],
            instruction=instruction,
        )
        stored_steps.append(
            {
                "phase_id": phase_id,
                "instruction": instruction,
                "prompt": prompt,
            }
        )

    return stored_steps


def _workflow_from_do_stored_steps(
    session: GoalSession,
    *,
    run_id: str,
    stored_steps: Sequence[dict[str, str]],
    checker_specs: Sequence[str],
) -> tuple[Workflow, list[tuple[str, str]]]:
    base_phases: list[Phase] = []
    step_instructions: list[tuple[str, str]] = []

    for index, stored_step in enumerate(stored_steps, start=1):
        phase_id = stored_step.get("phase_id")
        instruction = stored_step.get("instruction")
        prompt = stored_step.get("prompt")
        if not isinstance(phase_id, str) or not phase_id:
            raise JuvenalUsageError(f"Stored do() step {index} is missing phase_id")
        if not isinstance(instruction, str) or not instruction:
            raise JuvenalUsageError(f"Stored do() step {index} is missing instruction")
        if not isinstance(prompt, str) or not prompt:
            raise JuvenalUsageError(f"Stored do() step {index} is missing prompt")
        base_phases.append(Phase(id=phase_id, type="implement", prompt=prompt))
        step_instructions.append((phase_id, instruction))

    workflow = Workflow(
        name=f"do-{run_id}",
        phases=base_phases,
        backend=session.backend_name,
        working_dir=session.working_dir,
        max_bounces=session.max_bounces,
    )

    try:
        workflow = inject_checkers(workflow, checker_specs)
    except Exception as exc:
        raise JuvenalUsageError(str(exc)) from exc

    errors = validate_workflow(workflow)
    if errors:
        raise JuvenalUsageError(f"Embedded do() workflow validation failed: {'; '.join(errors)}")

    return workflow, step_instructions


def _collect_successful_do_history_entries(
    session: GoalSession,
    *,
    run_id: str,
    workflow: Workflow,
    engine: Engine,
    step_instructions: Sequence[tuple[str, str]],
) -> list[dict[str, Any]]:
    seen_phase_ids = {entry.get("phase_id") for entry in session.history}
    entries: list[dict[str, Any]] = []

    for phase_id, instruction in step_instructions:
        if phase_id in seen_phase_ids:
            continue
        if not _phase_completed_with_children(workflow, engine, phase_id):
            continue

        entries.append(
            {
                "kind": "do",
                "run_id": run_id,
                "phase_id": phase_id,
                "instruction": instruction,
                "summary": _build_history_summary(phase_id, instruction),
                "success": True,
            }
        )
        seen_phase_ids.add(phase_id)

    return entries


def _append_successful_do_history(
    session: GoalSession,
    *,
    stage_id: str | None,
    run_id: str,
    workflow: Workflow,
    engine: Engine,
    step_instructions: Sequence[tuple[str, str]],
) -> None:
    for history_entry in _collect_successful_do_history_entries(
        session,
        run_id=run_id,
        workflow=workflow,
        engine=engine,
        step_instructions=step_instructions,
    ):
        _persist_session_checkpoint(
            session,
            stage_id=stage_id,
            history_entries=[history_entry],
        )


def _staged_plan_owner_path(working_dir: Path) -> Path:
    return (working_dir / ".plan" / _STAGED_PLAN_OWNER_FILENAME).resolve()


def _planner_owner_identity(session: GoalSession, *, stage_id: str, run_id: str) -> dict[str, str]:
    if session.session_name is None:
        raise JuvenalUsageError("Staged juvenal.plan_and_do() requires a named session")
    return {
        "session_id": session.session_id,
        "session_name": session.session_name,
        "stage_id": stage_id,
        "run_id": run_id,
    }


def _planner_owner_conflict(path: Path, detail: str) -> JuvenalUsageError:
    return JuvenalUsageError(f"Invalid staged planner owner file at {path}: {detail}")


def _read_staged_plan_owner(path: Path) -> dict[str, str]:
    try:
        raw = json.loads(path.read_text(encoding="utf-8"))
    except FileNotFoundError as exc:
        raise _planner_owner_conflict(path, "missing owner file") from exc
    except (OSError, json.JSONDecodeError) as exc:
        raise _planner_owner_conflict(path, str(exc)) from exc

    if not isinstance(raw, dict):
        raise _planner_owner_conflict(path, "owner file must be a JSON object")

    owner: dict[str, str] = {}
    for key in ("session_id", "session_name", "stage_id", "run_id"):
        value = raw.get(key)
        if not isinstance(value, str) or not value:
            raise _planner_owner_conflict(path, f"owner field {key!r} must be a non-empty string")
        owner[key] = value

    return owner


def _write_staged_plan_owner(path: Path, owner: dict[str, str]) -> None:
    _write_json_atomic(path, owner)


def _collect_planner_asset_files(path: Path, seen_yaml: set[Path], assets: set[Path]) -> None:
    resolved = path.resolve()
    if resolved in seen_yaml:
        return
    seen_yaml.add(resolved)
    assets.add(resolved)

    raw = yaml.safe_load(resolved.read_text(encoding="utf-8")) or {}
    if not isinstance(raw, dict):
        raise JuvenalUsageError(f"Planner asset manifest source must be a mapping: {resolved}")

    includes = raw.get("include", [])
    if includes is None:
        includes = []
    if not isinstance(includes, list):
        raise JuvenalUsageError(f"Planner include list must be a sequence in {resolved}")
    for include_entry in includes:
        if not isinstance(include_entry, str):
            raise JuvenalUsageError(f"Planner include entries must be strings in {resolved}")
        _collect_planner_asset_files(resolved.parent / include_entry, seen_yaml, assets)

    def visit(node: Any) -> None:
        if isinstance(node, dict):
            prompt_file = node.get("prompt_file")
            if isinstance(prompt_file, str):
                assets.add((resolved.parent / prompt_file).resolve())
            for value in node.values():
                visit(value)
        elif isinstance(node, list):
            for value in node:
                visit(value)

    visit(raw)


def _build_planner_assets_manifest() -> dict[str, Any]:
    planner_yaml = (Path(__file__).parent / "workflows" / "plan.yaml").resolve()
    planner_validator = (Path(__file__).parent / "plan_validation.py").resolve()
    assets: set[Path] = {planner_validator}
    _collect_planner_asset_files(planner_yaml, set(), assets)

    files: list[dict[str, str]] = []
    digest_input_parts: list[str] = []
    for asset_path in sorted(assets):
        content = asset_path.read_bytes()
        file_digest = hashlib.sha256(content).hexdigest()
        files.append({"path": str(asset_path), "sha256": file_digest})
        digest_input_parts.append(f"{asset_path}\t{file_digest}")

    digest = hashlib.sha256("\n".join(digest_input_parts).encode("utf-8")).hexdigest()
    return {
        "root": str(planner_yaml),
        "files": files,
        "digest": digest,
    }


def _write_planner_assets_manifest(path: Path) -> dict[str, Any]:
    manifest = _build_planner_assets_manifest()
    _write_json_atomic(path, manifest)
    return manifest


def _load_planner_assets_manifest(run_id: str, path: Path) -> dict[str, Any]:
    try:
        raw = json.loads(path.read_text(encoding="utf-8"))
    except FileNotFoundError as exc:
        raise _raise_missing_stage_artifact(run_id, path, "Planner assets manifest is missing") from exc
    except (OSError, json.JSONDecodeError) as exc:
        raise _raise_missing_stage_artifact(run_id, path, f"Planner assets manifest is unreadable: {exc}") from exc

    if not isinstance(raw, dict) or not isinstance(raw.get("digest"), str):
        raise _raise_missing_stage_artifact(run_id, path, "Planner assets manifest is unreadable")
    return raw


def _ensure_planner_assets_unchanged(run_id: str, path: Path) -> None:
    recorded_manifest = _load_planner_assets_manifest(run_id, path)
    try:
        current_manifest = _build_planner_assets_manifest()
    except Exception as exc:
        raise _raise_missing_stage_artifact(run_id, path, f"Planner assets manifest is unreadable: {exc}") from exc

    if recorded_manifest.get("digest") != current_manifest["digest"]:
        raise _raise_missing_stage_artifact(run_id, path, "Planner assets changed since the stage was created")


def _rewind_planner_to_write_workflow(planner_state_path: Path, failure_context: str) -> None:
    planner_state = PipelineState.load(planner_state_path)
    planner_state.invalidate_from(_STAGED_PLAN_WRITE_WORKFLOW_PHASE_ID)
    target_attempt = planner_state.phases.get(_STAGED_PLAN_WRITE_WORKFLOW_PHASE_ID)
    planner_state.set_failure_context(
        _STAGED_PLAN_WRITE_WORKFLOW_PHASE_ID,
        failure_context,
        attempt=target_attempt.attempt if target_attempt is not None and target_attempt.attempt > 0 else None,
    )


def _prepare_plan_workspace(session: GoalSession, *, plan_dir: Path, workflow_path: Path) -> None:
    plan_dir.mkdir(exist_ok=True)
    git_context = _resolve_git_context(session.working_dir)
    if git_context is not None:
        _ensure_git_excluded(git_context, plan_dir)
        _ensure_git_excluded(git_context, workflow_path)


@contextmanager
def goal(
    description: str,
    *,
    working_dir: str | Path = ".",
    backend: str | Backend = "codex",
    max_bounces: int = 999,
    plain: bool = True,
    serialize: bool = False,
    clear_context_on_bounce: bool = False,
    artifact_dir: str | Path | None = None,
    session_name: str | None = None,
) -> Iterator[GoalSession]:
    """Create an embedded Juvenal session."""

    working_dir_path = _resolve_working_dir(working_dir)
    backend_name, backend_instance = _resolve_backend(backend)
    artifact_root = _resolve_artifact_root(working_dir_path, artifact_dir)

    git_context = _resolve_git_context(working_dir_path)
    if git_context is not None:
        _ensure_git_excluded(git_context, artifact_root)

    if session_name is None:
        session_id, session_artifact_dir = _resolve_anonymous_session_dir(artifact_root)
        manifest_path = _session_manifest_path(session_artifact_dir)
        session = _build_goal_session(
            goal_text=description,
            working_dir=working_dir_path,
            backend_name=backend_name,
            backend_instance=backend_instance,
            max_bounces=max_bounces,
            plain=plain,
            serialize=serialize,
            clear_context_on_bounce=clear_context_on_bounce,
            artifact_root=artifact_root,
            session_id=session_id,
            session_name=None,
            session_artifact_dir=session_artifact_dir,
            manifest_path=manifest_path,
        )
        _save_session_manifest(session)
    else:
        session_id, session_artifact_dir, existed = _resolve_named_session_dir(artifact_root, session_name)
        manifest_path = _session_manifest_path(session_artifact_dir)
        if existed:
            manifest = _load_session_manifest(manifest_path)
            _assert_session_identity_matches(
                session_name=session_id,
                manifest_path=manifest_path,
                manifest=manifest,
                goal_text=description,
                working_dir=working_dir_path,
                backend_name=backend_name,
                max_bounces=max_bounces,
                serialize=serialize,
                clear_context_on_bounce=clear_context_on_bounce,
            )
            session = _build_goal_session(
                goal_text=description,
                working_dir=working_dir_path,
                backend_name=backend_name,
                backend_instance=backend_instance,
                max_bounces=max_bounces,
                plain=plain,
                serialize=serialize,
                clear_context_on_bounce=clear_context_on_bounce,
                artifact_root=artifact_root,
                session_id=session_id,
                session_name=session_id,
                session_artifact_dir=session_artifact_dir,
                manifest_path=manifest_path,
                run_counter=manifest["run_counter"],
                history=manifest["history"],
                stages=manifest["stages"],
            )
        else:
            session = _build_goal_session(
                goal_text=description,
                working_dir=working_dir_path,
                backend_name=backend_name,
                backend_instance=backend_instance,
                max_bounces=max_bounces,
                plain=plain,
                serialize=serialize,
                clear_context_on_bounce=clear_context_on_bounce,
                artifact_root=artifact_root,
                session_id=session_id,
                session_name=session_id,
                session_artifact_dir=session_artifact_dir,
                manifest_path=manifest_path,
            )
            _save_session_manifest(session)

    token = _ACTIVE_SESSION.set(session)
    try:
        yield session
    finally:
        _ACTIVE_SESSION.reset(token)


def _run_do_one_shot(session: GoalSession, *, tasks: Sequence[str], checker_specs: Sequence[str]) -> None:
    run_id = _allocate_run_id(session)
    state_file = (session.session_artifact_dir / f"run-{run_id}-do.json").resolve()
    auto_stage_id = _stage_id_for_run("do", run_id)
    stored_steps = _build_do_stored_steps(session, run_id=run_id, tasks=tasks)
    _create_stage_record(
        session,
        stage_id=auto_stage_id,
        stage_record={
            "kind": "do",
            "run_id": run_id,
            "status": "running",
            "tasks": list(tasks),
            "checker_specs": list(checker_specs),
            "state_file": str(state_file),
            "stored_steps": stored_steps,
        },
    )

    try:
        workflow, step_instructions = _workflow_from_do_stored_steps(
            session,
            run_id=run_id,
            stored_steps=stored_steps,
            checker_specs=checker_specs,
        )
    except JuvenalUsageError:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise

    engine = Engine(
        workflow,
        backend_instance=session.backend_instance,
        state_file=state_file,
        plain=session.plain,
        clear_context_on_bounce=session.clear_context_on_bounce,
        serialize=session.serialize,
    )

    try:
        exit_code = engine.run()
    except Exception as exc:
        _append_successful_do_history(
            session,
            stage_id=auto_stage_id,
            run_id=run_id,
            workflow=workflow,
            engine=engine,
            step_instructions=step_instructions,
        )
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Embedded engine failed: {exc}",
            run_id=_run_label(run_id),
            inspection_path=state_file,
        ) from exc

    _append_successful_do_history(
        session,
        stage_id=auto_stage_id,
        run_id=run_id,
        workflow=workflow,
        engine=engine,
        step_instructions=step_instructions,
    )

    if exit_code != 0:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Embedded engine failed with exit code {exit_code}",
            run_id=_run_label(run_id),
            inspection_path=state_file,
        )

    _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "completed"})


def _run_do_staged(
    session: GoalSession,
    *,
    stage_id: str,
    tasks: Sequence[str],
    checker_specs: Sequence[str],
) -> None:
    validated_stage_id = _validate_stage_id(stage_id)
    expected_tasks = list(tasks)
    expected_checker_specs = list(checker_specs)
    stage_record = session.stages.get(validated_stage_id)
    resume = False

    if stage_record is None:
        run_id = _allocate_run_id(session)
        state_file = (session.session_artifact_dir / f"{_run_label(run_id)}-do.json").resolve()
        stored_steps = _build_do_stored_steps(session, run_id=run_id, tasks=tasks)
        _initialize_empty_pipeline_state(state_file)
        _create_stage_record(
            session,
            stage_id=validated_stage_id,
            stage_record={
                "kind": "do",
                "run_id": run_id,
                "status": "running",
                "tasks": expected_tasks,
                "checker_specs": expected_checker_specs,
                "state_file": str(state_file),
                "stored_steps": stored_steps,
            },
        )
        stage_record = session.stages[validated_stage_id]
    else:
        if stage_record.get("kind") != "do":
            raise JuvenalUsageError(
                f"Stage {validated_stage_id!r} already exists for kind {stage_record.get('kind')!r}, not 'do'"
            )
        if stage_record.get("tasks") != expected_tasks or stage_record.get("checker_specs") != expected_checker_specs:
            raise JuvenalUsageError(f"Stage {validated_stage_id!r} was created with different do() inputs")
        if stage_record.get("status") == "completed":
            return
        resume = True

    run_id = stage_record.get("run_id")
    if not isinstance(run_id, str) or not run_id:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} is missing run_id")
    state_file = _recorded_stage_path(validated_stage_id, stage_record, "state_file")
    stored_steps = stage_record.get("stored_steps")
    if not isinstance(stored_steps, list) or not stored_steps:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} is missing stored_steps")
    if resume:
        _ensure_required_state_file(
            run_id,
            state_file,
            detail="Recorded do() state file is missing or unreadable",
        )

    workflow, step_instructions = _workflow_from_do_stored_steps(
        session,
        run_id=run_id,
        stored_steps=stored_steps,
        checker_specs=checker_specs,
    )
    engine = Engine(
        workflow,
        backend_instance=session.backend_instance,
        resume=resume,
        state_file=state_file,
        plain=session.plain,
        clear_context_on_bounce=session.clear_context_on_bounce,
        serialize=session.serialize,
    )

    if resume:
        _append_successful_do_history(
            session,
            stage_id=validated_stage_id,
            run_id=run_id,
            workflow=workflow,
            engine=engine,
            step_instructions=step_instructions,
        )

    try:
        exit_code = engine.run()
    except Exception as exc:
        _append_successful_do_history(
            session,
            stage_id=validated_stage_id,
            run_id=run_id,
            workflow=workflow,
            engine=engine,
            step_instructions=step_instructions,
        )
        raise JuvenalExecutionError(
            f"Embedded engine failed: {exc}",
            run_id=_run_label(run_id),
            inspection_path=state_file,
        ) from exc

    _append_successful_do_history(
        session,
        stage_id=validated_stage_id,
        run_id=run_id,
        workflow=workflow,
        engine=engine,
        step_instructions=step_instructions,
    )

    if exit_code != 0:
        raise JuvenalExecutionError(
            f"Embedded engine failed with exit code {exit_code}",
            run_id=_run_label(run_id),
            inspection_path=state_file,
        )

    _persist_session_checkpoint(session, stage_id=validated_stage_id, stage_updates={"status": "completed"})


def do(
    task_or_tasks: str | Sequence[str],
    *,
    checker: str | None = None,
    checkers: Sequence[str] | None = None,
    stage_id: str | None = None,
) -> None:
    """Execute one or more embedded implement steps using the existing runtime."""

    session = _require_active_session("do")
    tasks = _normalize_do_tasks(task_or_tasks)
    checker_specs = _normalize_checker_specs(checker, checkers)

    try:
        for spec in checker_specs:
            parse_checker_string(spec)
    except Exception as exc:
        raise JuvenalUsageError(str(exc)) from exc

    if stage_id is None:
        _run_do_one_shot(session, tasks=tasks, checker_specs=checker_specs)
        return

    _run_do_staged(
        session,
        stage_id=stage_id,
        tasks=tasks,
        checker_specs=checker_specs,
    )


def _ensure_unstaged_plan_allowed(session: GoalSession) -> None:
    owner_path = _staged_plan_owner_path(session.working_dir)
    if owner_path.exists():
        raise JuvenalUsageError("Cannot mix staged and unstaged juvenal.plan_and_do() in the same workspace")
    if _persisted_staged_plan_owner(session) is not None:
        raise JuvenalUsageError("Cannot mix staged and unstaged juvenal.plan_and_do() in the same workspace")


def _persisted_staged_plan_owner(session: GoalSession) -> dict[str, Any] | None:
    for session_dir in sorted(session.artifact_root.iterdir()):
        manifest_path = _session_manifest_path(session_dir)
        if not manifest_path.exists():
            continue

        manifest = _load_session_manifest(manifest_path)
        if Path(manifest["working_dir"]).resolve() != session.working_dir:
            continue

        for stage_record in manifest["stages"].values():
            if stage_record.get("kind") != "plan-and-do":
                continue
            planner_owner = stage_record.get("planner_owner")
            if isinstance(planner_owner, dict):
                return planner_owner

    return None


def _persisted_plan_stage_conflict(session: GoalSession) -> JuvenalUsageError | None:
    for session_dir in sorted(session.artifact_root.iterdir()):
        manifest_path = _session_manifest_path(session_dir)
        if not manifest_path.exists():
            continue

        manifest = _load_session_manifest(manifest_path)
        if Path(manifest["working_dir"]).resolve() != session.working_dir:
            continue

        for stage_record in manifest["stages"].values():
            if stage_record.get("kind") != "plan-and-do":
                continue

            planner_owner = stage_record.get("planner_owner")
            if isinstance(planner_owner, dict):
                return JuvenalUsageError(
                    "Workspace already has a staged juvenal.plan_and_do() owner in the manifest: "
                    f"{planner_owner.get('session_name')!r} / {planner_owner.get('stage_id')!r}"
                )

            return JuvenalUsageError("Cannot mix staged and unstaged juvenal.plan_and_do() in the same workspace")

    return None


def _ensure_staged_plan_workspace_available(
    session: GoalSession,
    *,
    owner_path: Path,
    planner_state_path: Path,
) -> None:
    manifest_conflict = _persisted_plan_stage_conflict(session)
    if manifest_conflict is not None:
        raise manifest_conflict

    if owner_path.exists():
        owner = _read_staged_plan_owner(owner_path)
        raise JuvenalUsageError(
            "Workspace already has a staged juvenal.plan_and_do() owner: "
            f"{owner['session_name']!r} / {owner['stage_id']!r}"
        )
    if planner_state_path.exists():
        raise JuvenalUsageError(
            "Cannot start staged juvenal.plan_and_do() because planner state already exists without a staged owner"
        )


def _ensure_staged_plan_owner(expected_owner: dict[str, str], owner_path: Path) -> None:
    if owner_path.exists():
        actual_owner = _read_staged_plan_owner(owner_path)
        if actual_owner != expected_owner:
            raise JuvenalUsageError(
                "Workspace already has a different staged juvenal.plan_and_do() owner: "
                f"{actual_owner['session_name']!r} / {actual_owner['stage_id']!r}"
            )
        return
    _write_staged_plan_owner(owner_path, expected_owner)


def _load_planned_workflow_for_execution(
    session: GoalSession,
    *,
    run_id: str,
    workflow_path: Path,
) -> Workflow:
    try:
        workflow = load_workflow(workflow_path)
    except Exception as exc:
        raise JuvenalExecutionError(
            f"Planned workflow load failed: {exc}",
            run_id=_run_label(run_id),
            inspection_path=workflow_path,
        ) from exc

    workflow.working_dir = session.working_dir
    workflow.backend = session.backend_name
    workflow.max_bounces = session.max_bounces
    errors = validate_workflow(workflow)
    if errors:
        raise JuvenalExecutionError(
            f"Planned workflow validation failed: {'; '.join(errors)}",
            run_id=_run_label(run_id),
            inspection_path=workflow_path,
        )
    return workflow


def _finalize_staged_planning(
    session: GoalSession,
    *,
    stage_id: str,
    run_id: str,
    planner_state_path: Path,
    workflow_path: Path,
    archive_path: Path,
    planned_state_path: Path,
) -> Workflow:
    structure_path = planner_state_path.parent / "workflow-structure.yaml"

    try:
        raw_errors = validate_planned_workflow(structure_path, workflow_path)
    except Exception as exc:
        error = JuvenalExecutionError(
            f"Planned workflow load failed: {exc}",
            run_id=_run_label(run_id),
            inspection_path=structure_path if not structure_path.exists() else workflow_path,
        )
        _rewind_planner_to_write_workflow(planner_state_path, str(error))
        raise error from exc

    if raw_errors:
        error = JuvenalExecutionError(
            f"Planned workflow validation failed: {'; '.join(raw_errors)}",
            run_id=_run_label(run_id),
            inspection_path=workflow_path,
        )
        _rewind_planner_to_write_workflow(planner_state_path, str(error))
        raise error

    try:
        archive_path.write_bytes(workflow_path.read_bytes())
    except OSError as exc:
        error = JuvenalExecutionError(
            f"Failed to archive planned workflow: {exc}",
            run_id=_run_label(run_id),
            inspection_path=archive_path,
        )
        _rewind_planner_to_write_workflow(planner_state_path, str(error))
        raise error from exc

    try:
        _initialize_empty_pipeline_state(planned_state_path)
    except Exception as exc:
        error = JuvenalExecutionError(
            f"Failed to initialize planned execution state: {exc}",
            run_id=_run_label(run_id),
            inspection_path=planned_state_path,
        )
        _rewind_planner_to_write_workflow(planner_state_path, str(error))
        raise error from exc

    try:
        workflow = _load_planned_workflow_for_execution(
            session,
            run_id=run_id,
            workflow_path=archive_path,
        )
    except JuvenalExecutionError as exc:
        _rewind_planner_to_write_workflow(planner_state_path, str(exc))
        raise

    _persist_session_checkpoint(session, stage_id=stage_id, stage_updates={"status": "planner_complete"})
    return workflow


def _run_plan_and_do_one_shot(
    session: GoalSession,
    *,
    stripped_goal_text: str,
    normalized_goal_text: str,
) -> None:
    plan_dir = (session.working_dir / ".plan").resolve()
    workflow_path = (session.working_dir / "workflow.yaml").resolve()
    planner_state_path = (plan_dir / ".juvenal-state.json").resolve()
    run_id = _allocate_run_id(session)
    run_label = _run_label(run_id)
    archive_path = (session.session_artifact_dir / f"{run_label}-workflow.yaml").resolve()
    state_file = (session.session_artifact_dir / f"{run_label}-planned.json").resolve()
    auto_stage_id = _stage_id_for_run("plan-and-do", run_id)
    _create_stage_record(
        session,
        stage_id=auto_stage_id,
        stage_record={
            "kind": "plan-and-do",
            "run_id": run_id,
            "status": "running",
            "goal_text": stripped_goal_text,
            "workflow_archive": str(archive_path),
            "state_file": str(state_file),
            "planner_state_file": str(planner_state_path),
        },
    )

    composed_goal = _build_plan_and_do_goal(
        session=session,
        recent_history_summaries=_recent_history_summaries(session),
        goal_text=stripped_goal_text,
    )

    try:
        _prepare_plan_workspace(session, plan_dir=plan_dir, workflow_path=workflow_path)
        plan_result = _plan_workflow_internal(
            goal=composed_goal,
            project_dir=session.working_dir,
            backend_instance=session.backend_instance,
            plain=session.plain,
        )
    except Exception as exc:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Planning failed: {exc}",
            run_id=run_label,
            inspection_path=planner_state_path,
        ) from exc

    if not plan_result.success:
        detail = plan_result.error or "unknown planning failure"
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Planning failed: {detail}",
            run_id=run_label,
            inspection_path=planner_state_path,
        )

    try:
        archive_path.write_bytes(workflow_path.read_bytes())
        workflow = _load_planned_workflow_for_execution(
            session,
            run_id=run_id,
            workflow_path=workflow_path,
        )
    except JuvenalExecutionError:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise
    except Exception as exc:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Planned workflow load failed: {exc}",
            run_id=run_label,
            inspection_path=workflow_path,
        ) from exc

    engine = Engine(
        workflow,
        backend_instance=session.backend_instance,
        state_file=state_file,
        plain=session.plain,
        clear_context_on_bounce=session.clear_context_on_bounce,
        serialize=session.serialize,
    )

    try:
        exit_code = engine.run()
    except Exception as exc:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Planned engine failed: {exc}",
            run_id=run_label,
            inspection_path=state_file,
        ) from exc

    if exit_code != 0:
        _persist_session_checkpoint(session, stage_id=auto_stage_id, stage_updates={"status": "failed"})
        raise JuvenalExecutionError(
            f"Planned engine failed with exit code {exit_code}",
            run_id=run_label,
            inspection_path=state_file,
        )

    _persist_session_checkpoint(
        session,
        stage_id=auto_stage_id,
        stage_updates={"status": "completed"},
        history_entries=[
            {
                "kind": "plan_and_do",
                "run_id": run_id,
                "goal_text": stripped_goal_text,
                "summary": _normalize_history_summary(f"plan_and_do: {normalized_goal_text}"),
                "success": True,
            }
        ],
    )


def _run_plan_and_do_staged(
    session: GoalSession,
    *,
    stage_id: str,
    stripped_goal_text: str,
    normalized_goal_text: str,
) -> None:
    if session.session_name is None:
        raise JuvenalUsageError("Staged juvenal.plan_and_do() requires a named session")

    validated_stage_id = _validate_stage_id(stage_id)
    plan_dir = (session.working_dir / ".plan").resolve()
    workflow_path = (session.working_dir / "workflow.yaml").resolve()
    planner_state_path = (plan_dir / ".juvenal-state.json").resolve()
    owner_path = _staged_plan_owner_path(session.working_dir)
    stage_record = session.stages.get(validated_stage_id)
    workflow: Workflow | None = None
    planned_resume = False

    if stage_record is None:
        _prepare_plan_workspace(session, plan_dir=plan_dir, workflow_path=workflow_path)
        _ensure_staged_plan_workspace_available(
            session,
            owner_path=owner_path,
            planner_state_path=planner_state_path,
        )

        run_id = _allocate_run_id(session)
        run_label = _run_label(run_id)
        archive_path = (session.session_artifact_dir / f"{run_label}-workflow.yaml").resolve()
        planned_state_path = (session.session_artifact_dir / f"{run_label}-planned.json").resolve()
        planner_assets_path = (session.session_artifact_dir / f"{run_label}-planner-assets.json").resolve()
        planning_goal = _build_plan_and_do_goal(
            session=session,
            recent_history_summaries=_recent_history_summaries(session),
            goal_text=stripped_goal_text,
        )
        owner = _planner_owner_identity(session, stage_id=validated_stage_id, run_id=run_id)

        _initialize_empty_pipeline_state(planner_state_path)
        _write_planner_assets_manifest(planner_assets_path)
        _create_stage_record(
            session,
            stage_id=validated_stage_id,
            stage_record={
                "kind": "plan-and-do",
                "run_id": run_id,
                "status": "running",
                "goal_text": normalized_goal_text,
                "planning_goal": planning_goal,
                "planner_owner": owner,
                "planner_owner_path": str(owner_path),
                "planner_state_path": str(planner_state_path),
                "planner_assets_path": str(planner_assets_path),
                "workflow_archive_path": str(archive_path),
                "planned_state_path": str(planned_state_path),
            },
        )
        _write_staged_plan_owner(owner_path, owner)
        stage_record = session.stages[validated_stage_id]
        planning_resume = False
    else:
        if stage_record.get("kind") != "plan-and-do":
            raise JuvenalUsageError(
                f"Stage {validated_stage_id!r} already exists for kind {stage_record.get('kind')!r}, not 'plan-and-do'"
            )
        if stage_record.get("goal_text") != normalized_goal_text:
            raise JuvenalUsageError(f"Stage {validated_stage_id!r} was created with a different plan_and_do() goal")
        planning_resume = stage_record.get("status") == "running"

    run_id = stage_record.get("run_id")
    if not isinstance(run_id, str) or not run_id:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} is missing run_id")
    run_label = _run_label(run_id)
    planning_goal = stage_record.get("planning_goal")
    if not isinstance(planning_goal, str) or not planning_goal:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} is missing planning_goal")
    owner = stage_record.get("planner_owner")
    if not isinstance(owner, dict):
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} is missing planner_owner")
    expected_owner = _planner_owner_identity(session, stage_id=validated_stage_id, run_id=run_id)
    if owner != expected_owner:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} is owned by a different planner session")

    owner_path = _recorded_stage_path(validated_stage_id, stage_record, "planner_owner_path")
    planner_state_path = _recorded_stage_path(validated_stage_id, stage_record, "planner_state_path")
    planner_assets_path = _recorded_stage_path(validated_stage_id, stage_record, "planner_assets_path")
    archive_path = _recorded_stage_path(validated_stage_id, stage_record, "workflow_archive_path")
    planned_state_path = _recorded_stage_path(validated_stage_id, stage_record, "planned_state_path")
    stage_status = stage_record.get("status")

    if stage_status == "completed":
        _ensure_staged_plan_owner(expected_owner, owner_path)
        return

    if stage_status == "running":
        if planning_resume:
            _ensure_required_state_file(
                run_id,
                planner_state_path,
                detail="Recorded planner state file is missing or unreadable",
            )
            _ensure_planner_assets_unchanged(run_id, planner_assets_path)
        _ensure_staged_plan_owner(expected_owner, owner_path)
        _prepare_plan_workspace(session, plan_dir=plan_dir, workflow_path=workflow_path)

        try:
            plan_result = _plan_workflow_internal(
                goal=planning_goal,
                project_dir=session.working_dir,
                backend_instance=session.backend_instance,
                plain=session.plain,
                resume=planning_resume,
            )
        except Exception as exc:
            raise JuvenalExecutionError(
                f"Planning failed: {exc}",
                run_id=run_label,
                inspection_path=planner_state_path,
            ) from exc

        if not plan_result.success:
            detail = plan_result.error or "unknown planning failure"
            raise JuvenalExecutionError(
                f"Planning failed: {detail}",
                run_id=run_label,
                inspection_path=planner_state_path,
            )

        workflow = _finalize_staged_planning(
            session,
            stage_id=validated_stage_id,
            run_id=run_id,
            planner_state_path=planner_state_path,
            workflow_path=workflow_path,
            archive_path=archive_path,
            planned_state_path=planned_state_path,
        )
    elif stage_status == "planner_complete":
        _ensure_required_file_exists(
            run_id,
            archive_path,
            detail="Recorded archived workflow is missing or unreadable",
        )
        _ensure_required_state_file(
            run_id,
            planned_state_path,
            detail="Recorded planned execution state file is missing or unreadable",
        )
        _ensure_staged_plan_owner(expected_owner, owner_path)
        workflow = _load_planned_workflow_for_execution(
            session,
            run_id=run_id,
            workflow_path=archive_path,
        )
        planned_resume = True
    else:
        raise JuvenalUsageError(f"Stage {validated_stage_id!r} has unsupported status {stage_status!r}")

    engine = Engine(
        workflow,
        backend_instance=session.backend_instance,
        resume=planned_resume,
        state_file=planned_state_path,
        plain=session.plain,
        clear_context_on_bounce=session.clear_context_on_bounce,
        serialize=session.serialize,
    )

    try:
        exit_code = engine.run()
    except Exception as exc:
        raise JuvenalExecutionError(
            f"Planned engine failed: {exc}",
            run_id=run_label,
            inspection_path=planned_state_path,
        ) from exc

    if exit_code != 0:
        raise JuvenalExecutionError(
            f"Planned engine failed with exit code {exit_code}",
            run_id=run_label,
            inspection_path=planned_state_path,
        )

    _persist_session_checkpoint(
        session,
        stage_id=validated_stage_id,
        stage_updates={"status": "completed"},
        history_entries=[
            {
                "kind": "plan_and_do",
                "run_id": run_id,
                "goal_text": stripped_goal_text,
                "summary": _normalize_history_summary(f"plan_and_do: {normalized_goal_text}"),
                "success": True,
            }
        ],
    )


def plan_and_do(goal_text: str, *, stage_id: str | None = None) -> None:
    """Plan a workflow into the session working directory, then execute it as written."""

    session = _require_active_session("plan_and_do")
    stripped_goal_text, normalized_goal_text = _normalize_plan_goal_text(goal_text)

    if stage_id is None:
        _ensure_unstaged_plan_allowed(session)
        _run_plan_and_do_one_shot(
            session,
            stripped_goal_text=stripped_goal_text,
            normalized_goal_text=normalized_goal_text,
        )
        return

    _run_plan_and_do_staged(
        session,
        stage_id=stage_id,
        stripped_goal_text=stripped_goal_text,
        normalized_goal_text=normalized_goal_text,
    )


__all__ = [
    "GoalSession",
    "JuvenalExecutionError",
    "JuvenalUsageError",
    "do",
    "goal",
    "plan_and_do",
]



"""Unit tests for planner-specific workflow validation."""

from __future__ import annotations

import subprocess
import sys
from copy import deepcopy
from pathlib import Path

import yaml

from juvenal.plan_validation import validate_planned_workflow


def _base_structure() -> dict:
    return {
        "linear": True,
        "yaml_source_mode": "inline-only",
        "verifier_encoding": "explicit-phases",
        "phases": [
            {
                "order": 1,
                "id": "prepare",
                "type": "implement",
                "bounce_target": None,
                "required_preexisting_inputs": ["original/"],
            },
            {
                "order": 2,
                "id": "prepare-test",
                "type": "script",
                "bounce_target": "prepare",
                "required_preexisting_inputs": ["original/"],
            },
            {
                "order": 3,
                "id": "prepare-review",
                "type": "check",
                "bounce_target": "prepare",
                "required_preexisting_inputs": ["original/"],
            },
        ],
    }


def _base_workflow() -> dict:
    return {
        "name": "planned",
        "backend": "codex",
        "phases": [
            {
                "id": "prepare",
                "prompt": (
                    "Preexisting Inputs:\n- original/\n\n"
                    "New Outputs:\n- updated tree\n\n"
                    "Commit work to git before yielding."
                ),
            },
            {
                "id": "prepare-test",
                "type": "script",
                "run": "pytest -q",
                "bounce_target": "prepare",
            },
            {
                "id": "prepare-review",
                "type": "check",
                "bounce_target": "prepare",
                "prompt": "Role: Tester.\nRespond with VERDICT: PASS or VERDICT: FAIL: <reason>.",
            },
        ],
    }


def _smoke_structure() -> dict:
    return {
        "linear": True,
        "yaml_source_mode": "inline-only",
        "verifier_encoding": "explicit-phases",
        "required_preexisting_paths": ["original/"],
        "phases": [
            {"id": "analyze-prepared-inputs", "type": "implement"},
            {
                "id": "analyze-prepared-inputs-test",
                "type": "script",
                "bounce_target": "analyze-prepared-inputs",
            },
            {
                "id": "analyze-prepared-inputs-review",
                "type": "check",
                "bounce_target": "analyze-prepared-inputs",
            },
        ],
    }


def _smoke_workflow() -> dict:
    return {
        "name": "smoke",
        "backend": "codex",
        "phases": [
            {
                "id": "analyze-prepared-inputs",
                "prompt": "Inspect the prepared inputs and commit any changes before yielding.",
            },
            {
                "id": "analyze-prepared-inputs-test",
                "type": "script",
                "run": "true",
                "bounce_target": "analyze-prepared-inputs",
            },
            {
                "id": "analyze-prepared-inputs-review",
                "type": "check",
                "bounce_target": "analyze-prepared-inputs",
                "prompt": "Role: Tester.\nRespond with VERDICT: PASS or VERDICT: FAIL: <reason>.",
            },
        ],
    }


def _write_yaml(path: Path, data: dict) -> None:
    path.write_text(yaml.safe_dump(data, sort_keys=False), encoding="utf-8")


def _write_case(tmp_path: Path, structure: dict, workflow: dict) -> tuple[Path, Path]:
    structure_path = tmp_path / "workflow-structure.yaml"
    workflow_path = tmp_path / "workflow.yaml"
    _write_yaml(structure_path, structure)
    _write_yaml(workflow_path, workflow)
    return structure_path, workflow_path


def test_validate_planned_workflow_accepts_valid_linear_workflow(tmp_path):
    structure_path, workflow_path = _write_case(tmp_path, _base_structure(), _base_workflow())

    assert validate_planned_workflow(structure_path, workflow_path) == []


def test_plan_validation_module_entrypoint_accepts_valid_linear_workflow(tmp_path):
    structure_path, workflow_path = _write_case(tmp_path, _smoke_structure(), _smoke_workflow())

    result = subprocess.run(
        [sys.executable, "-m", "juvenal.plan_validation", str(structure_path), str(workflow_path)],
        capture_output=True,
        text=True,
        check=False,
    )

    assert result.returncode == 0
    assert result.stdout == ""


def test_validate_planned_workflow_rejects_order_id_type_mismatch(tmp_path):
    structure = _base_structure()
    workflow = _base_workflow()
    workflow["phases"][1]["id"] = "different"

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("does not match structure id" in error for error in errors)


def test_validate_planned_workflow_rejects_deferred_verifier(tmp_path):
    structure = {
        "linear": True,
        "yaml_source_mode": "inline-only",
        "verifier_encoding": "explicit-phases",
        "phases": [
            {
                "order": 1,
                "id": "prepare",
                "type": "implement",
                "bounce_target": None,
                "required_preexisting_inputs": ["original/"],
            },
            {
                "order": 2,
                "id": "port",
                "type": "implement",
                "bounce_target": None,
                "required_preexisting_inputs": ["original/"],
            },
            {
                "order": 3,
                "id": "prepare-review",
                "type": "check",
                "bounce_target": "prepare",
                "required_preexisting_inputs": ["original/"],
            },
        ],
    }
    workflow = {
        "name": "planned",
        "backend": "codex",
        "phases": [
            {"id": "prepare", "prompt": "Commit work to git before yielding."},
            {"id": "port", "prompt": "Commit work to git before yielding."},
            {
                "id": "prepare-review",
                "type": "check",
                "bounce_target": "prepare",
                "prompt": "Role: Tester.\nRespond with VERDICT: PASS or VERDICT: FAIL: <reason>.",
            },
        ],
    }

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("immediately preceding implement phase 'port'" in error for error in errors)


def test_validate_planned_workflow_rejects_missing_verifier_bounce_target(tmp_path):
    structure = _base_structure()
    workflow = _base_workflow()
    del workflow["phases"][2]["bounce_target"]

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("must define a single bounce_target" in error for error in errors)


def test_validate_planned_workflow_rejects_bounce_targets_lists(tmp_path):
    structure = _base_structure()
    workflow = _base_workflow()
    workflow["phases"][2]["bounce_targets"] = ["prepare"]

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("bounce_targets lists are not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_parallel_groups(tmp_path):
    structure = _base_structure()
    workflow = _base_workflow()
    workflow["parallel_groups"] = [{"phases": ["prepare", "prepare-review"]}]

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("parallel_groups is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_top_level_include(tmp_path):
    structure = _base_structure()
    workflow = _base_workflow()
    workflow["include"] = ["shared.yaml"]

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("top-level include is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_checks_indirection(tmp_path):
    structure = deepcopy(_base_structure())
    structure["phases"] = [structure["phases"][0]]
    workflow = {
        "name": "planned",
        "backend": "codex",
        "phases": [
            {
                "id": "prepare",
                "prompt": "Commit work to git before yielding.",
                "checks": [{"run": "pytest -q"}],
            }
        ],
    }

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("checks is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_top_level_checks_key(tmp_path):
    structure = deepcopy(_base_structure())
    structure["phases"] = [structure["phases"][0]]
    workflow = {
        "name": "planned",
        "backend": "codex",
        "checks": [{"run": "pytest -q"}],
        "phases": [
            {
                "id": "prepare",
                "prompt": "Commit work to git before yielding.",
            }
        ],
    }

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("top-level checks is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_checks_hidden_via_yaml_merge(tmp_path):
    structure = deepcopy(_base_structure())
    structure["phases"] = [structure["phases"][0]]
    workflow_path = tmp_path / "workflow.yaml"
    structure_path = tmp_path / "workflow-structure.yaml"
    _write_yaml(structure_path, structure)
    workflow_path.write_text(
        """\
name: planned
backend: codex
checker_defaults: &checker_defaults
  checks:
    - run: "pytest -q"
phases:
  - <<: *checker_defaults
    id: prepare
    prompt: "Commit work to git before yielding."
""",
        encoding="utf-8",
    )

    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("checks is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_phase_level_prompt_file(tmp_path):
    structure = deepcopy(_base_structure())
    structure["phases"] = [structure["phases"][0]]
    workflow = {
        "name": "planned",
        "backend": "codex",
        "phases": [
            {
                "id": "prepare",
                "prompt": "Commit work to git before yielding.",
                "prompt_file": "prompt.md",
            }
        ],
    }

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("prompt_file is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_phase_level_workflow_file(tmp_path):
    structure = deepcopy(_base_structure())
    structure["phases"] = [structure["phases"][0]]
    workflow = {
        "name": "planned",
        "backend": "codex",
        "phases": [
            {
                "id": "prepare",
                "prompt": "Commit work to git before yielding.",
                "workflow_file": "child.yaml",
            }
        ],
    }

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("workflow_file is not allowed" in error for error in errors)


def test_validate_planned_workflow_rejects_phase_level_workflow_dir(tmp_path):
    structure = deepcopy(_base_structure())
    structure["phases"] = [structure["phases"][0]]
    workflow = {
        "name": "planned",
        "backend": "codex",
        "phases": [
            {
                "id": "prepare",
                "prompt": "Commit work to git before yielding.",
                "workflow_dir": "sub-workflow",
            }
        ],
    }

    structure_path, workflow_path = _write_case(tmp_path, structure, workflow)
    errors = validate_planned_workflow(structure_path, workflow_path)

    assert any("workflow_dir is not allowed" in error for error in errors)



"""Tests for the toy embedded-API example module."""

from __future__ import annotations

from contextlib import contextmanager

from tests import mockup


def test_mockup_main_initializes_repo_before_goal_and_uses_expected_prompts(tmp_path, monkeypatch):
    tmpdir = tmp_path / "toy-example-repo"
    tmpdir.mkdir()
    goal_calls: list[dict[str, object]] = []
    do_calls: list[dict[str, object]] = []
    plan_calls: list[dict[str, object]] = []

    @contextmanager
    def fake_goal(description, *, working_dir, **kwargs):
        goal_calls.append({"description": description, "working_dir": working_dir, "kwargs": kwargs})
        yield object()

    def fake_do(task_or_tasks, **kwargs):
        do_calls.append({"task_or_tasks": task_or_tasks, "kwargs": kwargs})

    def fake_plan_and_do(prompt, **kwargs):
        plan_calls.append({"prompt": prompt, "kwargs": kwargs})

    monkeypatch.setattr(mockup.tempfile, "mkdtemp", lambda: str(tmpdir))
    monkeypatch.setattr(mockup.juvenal, "goal", fake_goal)
    monkeypatch.setattr(mockup.juvenal, "do", fake_do)
    monkeypatch.setattr(mockup.juvenal, "plan_and_do", fake_plan_and_do)

    returned = mockup.main()

    assert returned == str(tmpdir)
    assert len(goal_calls) == 1
    assert "toy todo CLI" in goal_calls[0]["description"]
    assert goal_calls[0]["working_dir"] == str(tmpdir)

    assert len(do_calls) == 3
    assert do_calls[0]["kwargs"] == {"checker": "pm"}
    assert do_calls[1]["kwargs"] == {"checkers": ["security-engineer"]}
    assert do_calls[2]["kwargs"] == {"checkers": ["tester", "senior-tester"]}
    assert [len(entry["task_or_tasks"]) if isinstance(entry["task_or_tasks"], list) else 1 for entry in do_calls] == [
        1,
        2,
        4,
    ]

    # first do: example brief
    example_brief_task = do_calls[0]["task_or_tasks"]
    assert isinstance(example_brief_task, str)
    assert f"{tmpdir}/example-brief.md" in example_brief_task
    assert "toy-todo" in example_brief_task
    assert "add" in example_brief_task
    assert "list" in example_brief_task
    assert "done" in example_brief_task
    assert "remove" in example_brief_task

    # second do: sample interactions
    sample_interaction_tasks = do_calls[1]["task_or_tasks"]
    assert isinstance(sample_interaction_tasks, list)
    assert len(sample_interaction_tasks) == 2
    assert f"{tmpdir}/sample-interactions.md" in sample_interaction_tasks[0]
    assert "edge-case" in sample_interaction_tasks[1]

    # third do: acceptance assets
    acceptance_asset_tasks = do_calls[2]["task_or_tasks"]
    assert isinstance(acceptance_asset_tasks, list)
    assert len(acceptance_asset_tasks) == 4
    assert f"{tmpdir}/acceptance-checklist.md" in acceptance_asset_tasks[0]
    assert f"{tmpdir}/smoke-test.sh" in acceptance_asset_tasks[3]

    # plan_and_do
    assert len(plan_calls) == 1
    final_prompt = plan_calls[0]["prompt"]
    assert "toy-todo" in final_prompt
    assert f"{tmpdir}/toy_app" in final_prompt
    assert "commit to git" in final_prompt
    assert "previous implementor" in final_prompt
    assert "already exist in" in final_prompt
    normalized = " ".join(final_prompt.split())
    assert "No generated phase may recreate those prep artifacts" in normalized
    assert "generated workflow must be linear" in final_prompt

    # no leftover libzstd references
    all_prompts: list[str] = []
    for entry in do_calls:
        if isinstance(entry["task_or_tasks"], list):
            all_prompts.extend(entry["task_or_tasks"])
        else:
            all_prompts.append(entry["task_or_tasks"])
    combined_text = "\n".join(all_prompts + [final_prompt])
    assert "libzstd" not in combined_text
    assert "Rust" not in combined_text



"""Planner-specific validation for generated workflow YAML."""

from __future__ import annotations

import argparse
from pathlib import Path
from typing import Any

import yaml

_VERIFIER_TYPES = {"check", "script"}
_ALLOWED_PHASE_TYPES = {"implement", "check", "script"}
_INLINE_ONLY_FORBIDDEN_PHASE_KEYS = ("prompt_file", "workflow_file", "workflow_dir")


def _read_yaml(path: Path, label: str) -> tuple[Any | None, list[str]]:
    try:
        raw = yaml.safe_load(path.read_text(encoding="utf-8"))
    except FileNotFoundError:
        return None, [f"{label} file not found: {path}"]
    except OSError as exc:
        return None, [f"{label} file is unreadable: {path}: {exc}"]
    except yaml.YAMLError as exc:
        return None, [f"{label} file is not valid YAML: {path}: {exc}"]
    return raw, []


def _phase_type(phase_data: dict[str, Any]) -> Any:
    return phase_data.get("type", "implement")


def _validate_required_input_list(value: Any, label: str, errors: list[str]) -> bool:
    if not isinstance(value, list) or any(not isinstance(item, str) for item in value):
        errors.append(f"{label} must be a list of strings")
        return False
    return True


def validate_planned_workflow(structure_path: Path, workflow_path: Path) -> list[str]:
    """Validate planner-produced workflow rules that generic workflow validation does not cover."""

    errors: list[str] = []

    structure_raw, structure_errors = _read_yaml(structure_path, "Structure")
    workflow_raw, workflow_errors = _read_yaml(workflow_path, "Workflow")
    errors.extend(structure_errors)
    errors.extend(workflow_errors)
    if structure_errors or workflow_errors:
        return errors

    if not isinstance(structure_raw, dict):
        return ["Structure file must be a YAML mapping"]
    if not isinstance(workflow_raw, dict):
        return ["Workflow file must be a YAML mapping"]

    structure_phases = structure_raw.get("phases")
    workflow_phases = workflow_raw.get("phases")
    if not isinstance(structure_phases, list):
        errors.append("Structure file must define phases as a list")
    if not isinstance(workflow_phases, list):
        errors.append("Workflow file must define phases as a list")
    if errors:
        return errors

    linear = structure_raw.get("linear")
    if linear is not True:
        errors.append("Structure file must set linear: true")

    yaml_source_mode = structure_raw.get("yaml_source_mode")
    if yaml_source_mode != "inline-only":
        errors.append("Structure file must set yaml_source_mode: inline-only")

    verifier_encoding = structure_raw.get("verifier_encoding")
    if verifier_encoding != "explicit-phases":
        errors.append("Structure file must set verifier_encoding: explicit-phases")

    has_global_required_inputs = False
    for key in ("required_preexisting_inputs", "required_preexisting_paths"):
        if key in structure_raw:
            has_global_required_inputs |= _validate_required_input_list(
                structure_raw.get(key),
                f"Structure field {key!r}",
                errors,
            )

    if linear is True and "parallel_groups" in workflow_raw:
        errors.append("parallel_groups is not allowed when linear: true")

    if yaml_source_mode == "inline-only" and "include" in workflow_raw:
        errors.append("top-level include is not allowed when yaml_source_mode is inline-only")
    if verifier_encoding == "explicit-phases" and "checks" in workflow_raw:
        errors.append("top-level checks is not allowed when verifier_encoding is explicit-phases")

    expected_count = len(structure_phases)
    actual_count = len(workflow_phases)
    if actual_count != expected_count:
        errors.append(f"Workflow phase count {actual_count} does not match structure phase count {expected_count}")

    shared_count = min(expected_count, actual_count)
    for index in range(shared_count):
        structure_phase = structure_phases[index]
        workflow_phase = workflow_phases[index]
        phase_number = index + 1

        if not isinstance(structure_phase, dict):
            errors.append(f"Structure phase {phase_number} must be a YAML mapping")
            continue
        if not isinstance(workflow_phase, dict):
            errors.append(f"Workflow phase {phase_number} must be a YAML mapping")
            continue

        order = structure_phase.get("order")
        if "order" in structure_phase and order != phase_number:
            errors.append(f"Structure phase {phase_number} must set order: {phase_number}")

        structure_id = structure_phase.get("id")
        workflow_id = workflow_phase.get("id")
        if not isinstance(structure_id, str) or not structure_id:
            errors.append(f"Structure phase {phase_number} must define a non-empty id")
        if structure_id != workflow_id:
            errors.append(
                f"Workflow phase {phase_number} id {workflow_id!r} does not match structure id {structure_id!r}"
            )

        structure_type = structure_phase.get("type")
        workflow_type = _phase_type(workflow_phase)
        if structure_type not in _ALLOWED_PHASE_TYPES:
            errors.append(
                f"Structure phase {structure_id!r}: invalid type {structure_type!r} "
                f"(must be one of {sorted(_ALLOWED_PHASE_TYPES)})"
            )
        if workflow_type != structure_type:
            errors.append(
                f"Workflow phase {workflow_id!r}: type {workflow_type!r} does not match structure type "
                f"{structure_type!r}"
            )

        if "required_preexisting_inputs" in structure_phase:
            _validate_required_input_list(
                structure_phase.get("required_preexisting_inputs"),
                f"Structure phase {structure_id or phase_number!r}: required_preexisting_inputs",
                errors,
            )
        elif not has_global_required_inputs:
            errors.append(
                "Structure file must declare required preexisting inputs either via top-level "
                "required_preexisting_inputs/required_preexisting_paths or per-phase "
                "required_preexisting_inputs"
            )

        structure_bounce = structure_phase.get("bounce_target")
        if (
            "bounce_target" in structure_phase
            and structure_bounce is not None
            and not isinstance(structure_bounce, str)
        ):
            errors.append(f"Structure phase {structure_id!r}: bounce_target must be a string or null")

        workflow_bounce = workflow_phase.get("bounce_target")
        if workflow_bounce != structure_bounce:
            errors.append(
                f"Workflow phase {workflow_id!r}: bounce_target {workflow_bounce!r} does not match structure "
                f"bounce_target {structure_bounce!r}"
            )

        if linear is True and "bounce_targets" in workflow_phase:
            errors.append(f"Workflow phase {workflow_id!r}: bounce_targets lists are not allowed when linear: true")

        if verifier_encoding == "explicit-phases" and "checks" in workflow_phase:
            errors.append(
                f"Workflow phase {workflow_id!r}: checks is not allowed when verifier_encoding is explicit-phases"
            )

        if yaml_source_mode == "inline-only":
            for forbidden_key in _INLINE_ONLY_FORBIDDEN_PHASE_KEYS:
                if forbidden_key in workflow_phase:
                    errors.append(
                        f"Workflow phase {workflow_id!r}: {forbidden_key} is not allowed when "
                        "yaml_source_mode is inline-only"
                    )

    last_implement_id: str | None = None
    for phase_data in workflow_phases:
        if not isinstance(phase_data, dict):
            continue
        phase_id = phase_data.get("id", "<unknown>")
        phase_type = _phase_type(phase_data)
        if phase_type == "implement":
            last_implement_id = phase_id if isinstance(phase_id, str) and phase_id else None
            continue
        if phase_type not in _VERIFIER_TYPES:
            continue
        bounce_target = phase_data.get("bounce_target")
        if not isinstance(bounce_target, str) or not bounce_target:
            errors.append(f"Workflow verifier phase {phase_id!r} must define a single bounce_target")
        if last_implement_id is None:
            errors.append(f"Workflow verifier phase {phase_id!r} must follow an implement phase")
        elif bounce_target != last_implement_id:
            errors.append(
                f"Workflow verifier phase {phase_id!r} must bounce to the immediately preceding implement phase "
                f"{last_implement_id!r}, got {bounce_target!r}"
            )

    return errors


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(description="Validate planner-generated workflow structure and topology.")
    parser.add_argument("structure", help="Path to .plan/workflow-structure.yaml")
    parser.add_argument("workflow", help="Path to generated workflow.yaml")
    args = parser.parse_args(argv)

    errors = validate_planned_workflow(Path(args.structure), Path(args.workflow))
    if errors:
        for error in errors:
            print(error)
        return 1
    return 0


if __name__ == "__main__":
    raise SystemExit(main())
